from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, Optional, Tuple

from xsdata.models.datatype import XmlDateTime

from ..w3c.xmldsig_core import (
    CanonicalizationMethod,
    DigestMethod,
    DigestValue,
    Signature,
    Transforms,
    X509IssuerSerialType,
)

__NAMESPACE__ = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class AnyType:
    any_attributes: Dict[str, str] = field(
        default_factory=dict,
        metadata={
            "type": "Attributes",
            "namespace": "##any",
        },
    )
    content: Tuple[object, ...] = field(
        default_factory=tuple,
        metadata={
            "type": "Wildcard",
            "namespace": "##any",
            "mixed": True,
        },
    )


@dataclass(frozen=True)
class CRLIdentifierType:
    issuer: Optional[str] = field(
        default=None,
        metadata={
            "name": "Issuer",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )
    issue_time: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "name": "IssueTime",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )
    number: Optional[int] = field(
        default=None,
        metadata={
            "name": "Number",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    uri: Optional[str] = field(
        default=None,
        metadata={
            "name": "URI",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class DocumentationReferencesType:
    documentation_reference: Tuple[str, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "DocumentationReference",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class EncapsulatedPKIDataType:
    value: Optional[bytes] = field(
        default=None,
        metadata={
            "required": True,
            "format": "base64",
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )
    encoding: Optional[str] = field(
        default=None,
        metadata={
            "name": "Encoding",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class IncludeType:
    uri: Optional[str] = field(
        default=None,
        metadata={
            "name": "URI",
            "type": "Attribute",
            "required": True,
        },
    )
    referenced_data: Optional[bool] = field(
        default=None,
        metadata={
            "name": "referencedData",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class IntegerListType:
    int_value: Tuple[int, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "int",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )


class QualifierType(Enum):
    OIDAS_URI = "OIDAsURI"
    OIDAS_URN = "OIDAsURN"


@dataclass(frozen=True)
class QualifyingPropertiesReferenceType:
    uri: Optional[str] = field(
        default=None,
        metadata={
            "name": "URI",
            "type": "Attribute",
            "required": True,
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class ResponderIDType:
    by_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "ByName",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    by_key: Optional[bytes] = field(
        default=None,
        metadata={
            "name": "ByKey",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "format": "base64",
        },
    )


@dataclass(frozen=True)
class SPURI:
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"

    value: str = field(
        default="",
        metadata={
            "required": True,
        },
    )


@dataclass(frozen=True)
class SignatureProductionPlaceType:
    city: Optional[str] = field(
        default=None,
        metadata={
            "name": "City",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    state_or_province: Optional[str] = field(
        default=None,
        metadata={
            "name": "StateOrProvince",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    postal_code: Optional[str] = field(
        default=None,
        metadata={
            "name": "PostalCode",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    country_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "CountryName",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )


@dataclass(frozen=True)
class SigningTime:
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"

    value: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "required": True,
        },
    )


@dataclass(frozen=True)
class Anytype(AnyType):
    class Meta:
        name = "Any"
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class CRLValuesType:
    encapsulated_crlvalue: Tuple[EncapsulatedPKIDataType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "EncapsulatedCRLValue",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class CertificateValuesType:
    encapsulated_x509_certificate: Tuple[EncapsulatedPKIDataType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "EncapsulatedX509Certificate",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    other_certificate: Tuple[AnyType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "OtherCertificate",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class CertifiedRolesListType:
    certified_role: Tuple[EncapsulatedPKIDataType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "CertifiedRole",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class ClaimedRolesListType:
    claimed_role: Tuple[AnyType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "ClaimedRole",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class CommitmentTypeQualifiersListType:
    commitment_type_qualifier: Tuple[AnyType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "CommitmentTypeQualifier",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )


@dataclass(frozen=True)
class CounterSignatureType:
    signature: Optional[Signature] = field(
        default=None,
        metadata={
            "name": "Signature",
            "type": "Element",
            "namespace": "http://www.w3.org/2000/09/xmldsig#",
            "required": True,
        },
    )


@dataclass(frozen=True)
class DigestAlgAndValueType:
    digest_method: Optional[DigestMethod] = field(
        default=None,
        metadata={
            "name": "DigestMethod",
            "type": "Element",
            "namespace": "http://www.w3.org/2000/09/xmldsig#",
            "required": True,
        },
    )
    digest_value: Optional[DigestValue] = field(
        default=None,
        metadata={
            "name": "DigestValue",
            "type": "Element",
            "namespace": "http://www.w3.org/2000/09/xmldsig#",
            "required": True,
        },
    )


@dataclass(frozen=True)
class EncapsulatedPKIData(EncapsulatedPKIDataType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class IdentifierType:
    value: str = field(
        default="",
        metadata={
            "required": True,
        },
    )
    qualifier: Optional[QualifierType] = field(
        default=None,
        metadata={
            "name": "Qualifier",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class Include(IncludeType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class NoticeReferenceType:
    organization: Optional[str] = field(
        default=None,
        metadata={
            "name": "Organization",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )
    notice_numbers: Optional[IntegerListType] = field(
        default=None,
        metadata={
            "name": "NoticeNumbers",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )


@dataclass(frozen=True)
class OCSPIdentifierType:
    responder_id: Optional[ResponderIDType] = field(
        default=None,
        metadata={
            "name": "ResponderID",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )
    produced_at: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "name": "ProducedAt",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )
    uri: Optional[str] = field(
        default=None,
        metadata={
            "name": "URI",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class OCSPValuesType:
    encapsulated_ocspvalue: Tuple[EncapsulatedPKIDataType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "EncapsulatedOCSPValue",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class OtherCertStatusRefsType:
    other_ref: Tuple[AnyType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "OtherRef",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class OtherCertStatusValuesType:
    other_value: Tuple[AnyType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "OtherValue",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class QualifyingPropertiesReference(QualifyingPropertiesReferenceType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class ReferenceInfoType:
    digest_method: Optional[DigestMethod] = field(
        default=None,
        metadata={
            "name": "DigestMethod",
            "type": "Element",
            "namespace": "http://www.w3.org/2000/09/xmldsig#",
            "required": True,
        },
    )
    digest_value: Optional[DigestValue] = field(
        default=None,
        metadata={
            "name": "DigestValue",
            "type": "Element",
            "namespace": "http://www.w3.org/2000/09/xmldsig#",
            "required": True,
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )
    uri: Optional[str] = field(
        default=None,
        metadata={
            "name": "URI",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class SigPolicyQualifiersListType:
    sig_policy_qualifier: Tuple[AnyType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "SigPolicyQualifier",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class SignatureProductionPlace(SignatureProductionPlaceType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class UnsignedDataObjectPropertiesType:
    unsigned_data_object_property: Tuple[AnyType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "UnsignedDataObjectProperty",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "min_occurs": 1,
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class AttrAuthoritiesCertValues(CertificateValuesType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class CRLRefType:
    digest_alg_and_value: Optional[DigestAlgAndValueType] = field(
        default=None,
        metadata={
            "name": "DigestAlgAndValue",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )
    crlidentifier: Optional[CRLIdentifierType] = field(
        default=None,
        metadata={
            "name": "CRLIdentifier",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )


@dataclass(frozen=True)
class CertIDType:
    cert_digest: Optional[DigestAlgAndValueType] = field(
        default=None,
        metadata={
            "name": "CertDigest",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )
    issuer_serial: Optional[X509IssuerSerialType] = field(
        default=None,
        metadata={
            "name": "IssuerSerial",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )
    uri: Optional[str] = field(
        default=None,
        metadata={
            "name": "URI",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class CertificateValues(CertificateValuesType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class CounterSignature(CounterSignatureType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class OCSPRefType:
    ocspidentifier: Optional[OCSPIdentifierType] = field(
        default=None,
        metadata={
            "name": "OCSPIdentifier",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )
    digest_alg_and_value: Optional[DigestAlgAndValueType] = field(
        default=None,
        metadata={
            "name": "DigestAlgAndValue",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )


@dataclass(frozen=True)
class ObjectIdentifierType:
    identifier: Optional[IdentifierType] = field(
        default=None,
        metadata={
            "name": "Identifier",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )
    description: Optional[str] = field(
        default=None,
        metadata={
            "name": "Description",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    documentation_references: Optional[DocumentationReferencesType] = field(
        default=None,
        metadata={
            "name": "DocumentationReferences",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )


@dataclass(frozen=True)
class ReferenceInfo(ReferenceInfoType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class RevocationValuesType:
    crlvalues: Optional[CRLValuesType] = field(
        default=None,
        metadata={
            "name": "CRLValues",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    ocspvalues: Optional[OCSPValuesType] = field(
        default=None,
        metadata={
            "name": "OCSPValues",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    other_values: Optional[OtherCertStatusValuesType] = field(
        default=None,
        metadata={
            "name": "OtherValues",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class SPUserNoticeType:
    notice_ref: Optional[NoticeReferenceType] = field(
        default=None,
        metadata={
            "name": "NoticeRef",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    explicit_text: Optional[str] = field(
        default=None,
        metadata={
            "name": "ExplicitText",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )


@dataclass(frozen=True)
class SignerRoleType:
    claimed_roles: Optional[ClaimedRolesListType] = field(
        default=None,
        metadata={
            "name": "ClaimedRoles",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    certified_roles: Optional[CertifiedRolesListType] = field(
        default=None,
        metadata={
            "name": "CertifiedRoles",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )


@dataclass(frozen=True)
class UnsignedDataObjectProperties(UnsignedDataObjectPropertiesType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class AttributeRevocationValues(RevocationValuesType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class CRLRefsType:
    crlref: Tuple[CRLRefType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "CRLRef",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class CertIDListType:
    cert: Tuple[CertIDType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "Cert",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class CommitmentTypeIndicationType:
    commitment_type_id: Optional[ObjectIdentifierType] = field(
        default=None,
        metadata={
            "name": "CommitmentTypeId",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )
    object_reference: Tuple[str, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "ObjectReference",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    all_signed_data_objects: Optional[object] = field(
        default=None,
        metadata={
            "name": "AllSignedDataObjects",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    commitment_type_qualifiers: Optional[CommitmentTypeQualifiersListType] = (
        field(
            default=None,
            metadata={
                "name": "CommitmentTypeQualifiers",
                "type": "Element",
                "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            },
        )
    )


@dataclass(frozen=True)
class DataObjectFormatType:
    description: Optional[str] = field(
        default=None,
        metadata={
            "name": "Description",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    object_identifier: Optional[ObjectIdentifierType] = field(
        default=None,
        metadata={
            "name": "ObjectIdentifier",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    mime_type: Optional[str] = field(
        default=None,
        metadata={
            "name": "MimeType",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    encoding: Optional[str] = field(
        default=None,
        metadata={
            "name": "Encoding",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    object_reference: Optional[str] = field(
        default=None,
        metadata={
            "name": "ObjectReference",
            "type": "Attribute",
            "required": True,
        },
    )


@dataclass(frozen=True)
class GenericTimeStampType:
    include: Tuple[Include, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "Include",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    reference_info: Tuple[ReferenceInfo, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "ReferenceInfo",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    canonicalization_method: Optional[CanonicalizationMethod] = field(
        default=None,
        metadata={
            "name": "CanonicalizationMethod",
            "type": "Element",
            "namespace": "http://www.w3.org/2000/09/xmldsig#",
        },
    )
    encapsulated_time_stamp: Tuple[EncapsulatedPKIDataType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "EncapsulatedTimeStamp",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    xmltime_stamp: Tuple[AnyType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "XMLTimeStamp",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class OCSPRefsType:
    ocspref: Tuple[OCSPRefType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "OCSPRef",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class ObjectIdentifier(ObjectIdentifierType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class RevocationValues(RevocationValuesType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class SPUserNotice(SPUserNoticeType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class SignaturePolicyIdType:
    sig_policy_id: Optional[ObjectIdentifierType] = field(
        default=None,
        metadata={
            "name": "SigPolicyId",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )
    transforms: Optional[Transforms] = field(
        default=None,
        metadata={
            "name": "Transforms",
            "type": "Element",
            "namespace": "http://www.w3.org/2000/09/xmldsig#",
        },
    )
    sig_policy_hash: Optional[DigestAlgAndValueType] = field(
        default=None,
        metadata={
            "name": "SigPolicyHash",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )
    sig_policy_qualifiers: Optional[SigPolicyQualifiersListType] = field(
        default=None,
        metadata={
            "name": "SigPolicyQualifiers",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )


@dataclass(frozen=True)
class SignerRole(SignerRoleType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class CommitmentTypeIndication(CommitmentTypeIndicationType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class CompleteCertificateRefsType:
    cert_refs: Optional[CertIDListType] = field(
        default=None,
        metadata={
            "name": "CertRefs",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "required": True,
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class CompleteRevocationRefsType:
    crlrefs: Optional[CRLRefsType] = field(
        default=None,
        metadata={
            "name": "CRLRefs",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    ocsprefs: Optional[OCSPRefsType] = field(
        default=None,
        metadata={
            "name": "OCSPRefs",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    other_refs: Optional[OtherCertStatusRefsType] = field(
        default=None,
        metadata={
            "name": "OtherRefs",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class DataObjectFormat(DataObjectFormatType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class OtherTimeStampType(GenericTimeStampType):
    include: Any = field(
        init=False,
        metadata={
            "type": "Ignore",
        },
    )
    reference_info: Tuple[ReferenceInfo, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "ReferenceInfo",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class SignaturePolicyIdentifierType:
    signature_policy_id: Optional[SignaturePolicyIdType] = field(
        default=None,
        metadata={
            "name": "SignaturePolicyId",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    signature_policy_implied: Optional[object] = field(
        default=None,
        metadata={
            "name": "SignaturePolicyImplied",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )


@dataclass(frozen=True)
class SigningCertificate(CertIDListType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class XAdESTimeStampType(GenericTimeStampType):
    reference_info: Any = field(
        init=False,
        metadata={
            "type": "Ignore",
        },
    )


@dataclass(frozen=True)
class AllDataObjectsTimeStamp(XAdESTimeStampType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class ArchiveTimeStamp(XAdESTimeStampType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class AttributeCertificateRefs(CompleteCertificateRefsType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class AttributeRevocationRefs(CompleteRevocationRefsType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class CompleteCertificateRefs(CompleteCertificateRefsType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class CompleteRevocationRefs(CompleteRevocationRefsType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class IndividualDataObjectsTimeStamp(XAdESTimeStampType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class OtherTimeStamp(OtherTimeStampType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class RefsOnlyTimeStamp(XAdESTimeStampType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class SigAndRefsTimeStamp(XAdESTimeStampType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class SignaturePolicyIdentifier(SignaturePolicyIdentifierType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class SignatureTimeStamp(XAdESTimeStampType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class SignedDataObjectPropertiesType:
    data_object_format: Tuple[DataObjectFormatType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "DataObjectFormat",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    commitment_type_indication: Tuple[CommitmentTypeIndicationType, ...] = (
        field(
            default_factory=tuple,
            metadata={
                "name": "CommitmentTypeIndication",
                "type": "Element",
                "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            },
        )
    )
    all_data_objects_time_stamp: Tuple[XAdESTimeStampType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "AllDataObjectsTimeStamp",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    individual_data_objects_time_stamp: Tuple[XAdESTimeStampType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "IndividualDataObjectsTimeStamp",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class SignedSignaturePropertiesType:
    signing_time: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "name": "SigningTime",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    signing_certificate: Optional[CertIDListType] = field(
        default=None,
        metadata={
            "name": "SigningCertificate",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    signature_policy_identifier: Optional[SignaturePolicyIdentifierType] = (
        field(
            default=None,
            metadata={
                "name": "SignaturePolicyIdentifier",
                "type": "Element",
                "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            },
        )
    )
    signature_production_place: Optional[SignatureProductionPlaceType] = field(
        default=None,
        metadata={
            "name": "SignatureProductionPlace",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    signer_role: Optional[SignerRoleType] = field(
        default=None,
        metadata={
            "name": "SignerRole",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class UnsignedSignaturePropertiesType:
    counter_signature: Tuple[CounterSignatureType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "CounterSignature",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    signature_time_stamp: Tuple[XAdESTimeStampType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "SignatureTimeStamp",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    complete_certificate_refs: Tuple[CompleteCertificateRefsType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "CompleteCertificateRefs",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    complete_revocation_refs: Tuple[CompleteRevocationRefsType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "CompleteRevocationRefs",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    attribute_certificate_refs: Tuple[CompleteCertificateRefsType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "AttributeCertificateRefs",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    attribute_revocation_refs: Tuple[CompleteRevocationRefsType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "AttributeRevocationRefs",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    sig_and_refs_time_stamp: Tuple[XAdESTimeStampType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "SigAndRefsTimeStamp",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    refs_only_time_stamp: Tuple[XAdESTimeStampType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "RefsOnlyTimeStamp",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    certificate_values: Tuple[CertificateValuesType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "CertificateValues",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    revocation_values: Tuple[RevocationValuesType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "RevocationValues",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    attr_authorities_cert_values: Tuple[CertificateValuesType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "AttrAuthoritiesCertValues",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    attribute_revocation_values: Tuple[RevocationValuesType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "AttributeRevocationValues",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    archive_time_stamp: Tuple[XAdESTimeStampType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "ArchiveTimeStamp",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    other_element: Tuple[object, ...] = field(
        default_factory=tuple,
        metadata={
            "type": "Wildcard",
            "namespace": "##other",
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class XAdESTimeStamp(XAdESTimeStampType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class SignedDataObjectProperties(SignedDataObjectPropertiesType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class SignedPropertiesType:
    signed_signature_properties: Optional[SignedSignaturePropertiesType] = (
        field(
            default=None,
            metadata={
                "name": "SignedSignatureProperties",
                "type": "Element",
                "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            },
        )
    )
    signed_data_object_properties: Optional[SignedDataObjectPropertiesType] = (
        field(
            default=None,
            metadata={
                "name": "SignedDataObjectProperties",
                "type": "Element",
                "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            },
        )
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class SignedSignatureProperties(SignedSignaturePropertiesType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class UnsignedPropertiesType:
    unsigned_signature_properties: Optional[UnsignedSignaturePropertiesType] = (
        field(
            default=None,
            metadata={
                "name": "UnsignedSignatureProperties",
                "type": "Element",
                "namespace": "http://uri.etsi.org/01903/v1.3.2#",
            },
        )
    )
    unsigned_data_object_properties: Optional[
        UnsignedDataObjectPropertiesType
    ] = field(
        default=None,
        metadata={
            "name": "UnsignedDataObjectProperties",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class UnsignedSignatureProperties(UnsignedSignaturePropertiesType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class QualifyingPropertiesType:
    signed_properties: Optional[SignedPropertiesType] = field(
        default=None,
        metadata={
            "name": "SignedProperties",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    unsigned_properties: Optional[UnsignedPropertiesType] = field(
        default=None,
        metadata={
            "name": "UnsignedProperties",
            "type": "Element",
            "namespace": "http://uri.etsi.org/01903/v1.3.2#",
        },
    )
    target: Optional[str] = field(
        default=None,
        metadata={
            "name": "Target",
            "type": "Attribute",
            "required": True,
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "name": "Id",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class SignedProperties(SignedPropertiesType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class UnsignedProperties(UnsignedPropertiesType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"


@dataclass(frozen=True)
class QualifyingProperties(QualifyingPropertiesType):
    class Meta:
        namespace = "http://uri.etsi.org/01903/v1.3.2#"



"""
This module contains utilities for allowing dataclasses to be populated by
user-provided configuration (e.g. from a Yaml file).

.. note::
    On naming conventions: this module converts hyphens in key names to
    underscores as a matter of course.
"""

import dataclasses
import re
from typing import Optional, Set, Type, Union

from asn1crypto.core import BitString, ObjectIdentifier

from .errors import ConfigurationError

__all__ = [
    'ConfigurableMixin',
    'check_config_keys',
    'OID_REGEX',
    'process_oid',
    'process_oids',
    'process_bit_string_flags',
]

_noneType = type(None)


def _unwrap_type_annot(thing) -> Optional[type]:
    if isinstance(thing, type):
        the_type = thing
    else:
        from typing import get_args, get_origin

        # is it an optional? (i.e. Union[X, None])
        # if so, retrieve the wrapped type
        if get_origin(thing) is not Union:
            return None
        try:
            type1, type2 = get_args(thing)
            if type2 is not _noneType:
                return None
        except (ValueError, TypeError):
            return None
        the_type = type1
    return the_type if isinstance(the_type, type) else None


def _has_default(f: dataclasses.Field):
    return (
        f.default_factory is not dataclasses.MISSING
        or f.default is not dataclasses.MISSING
    )


@dataclasses.dataclass(frozen=True)
class ConfigurableMixin:
    """General configuration mixin for dataclasses"""

    @classmethod
    def process_entries(cls, config_dict):
        """
        Hook method that can modify the configuration dictionary
        to overwrite or tweak some of their values (e.g. to convert string
        parameters into more complex Python objects)

        Subclasses that override this method should call
        ``super().process_entries()``, and leave keys that they do not
        recognise untouched.

        :param config_dict:
            A dictionary containing configuration values.
        :raises ConfigurationError:
            when there is a problem processing a relevant entry.
        """
        pass

    @classmethod
    def _process_configurable_fields(cls, config_dict):
        # automatically parse values for configurable fields
        for f in dataclasses.fields(cls):
            field_type = _unwrap_type_annot(f.type)
            if field_type is None or not issubclass(
                field_type, ConfigurableMixin
            ):
                continue
            # if the field has a value in the config dict, attempt to parse it
            try:
                field_config_dict = config_dict[f.name]
            except KeyError:
                continue
            try:
                field_value = field_type.from_config(field_config_dict)
            except ConfigurationError as e:
                raise ConfigurationError(
                    f"Error while processing configurable field '{f.name}': "
                    f"{repr(e)}"
                ) from e
            config_dict[f.name] = field_value

    @classmethod
    def check_config_keys(cls, keys_supplied: Set[str]):
        """
        Check whether all supplied keys are meaningful.

        :param keys_supplied:
            The keys supplied in the configuration.
        :raises ConfigurationError: if at least one key does not make sense.
        """

        check_config_keys(
            cls.__name__,
            {f.name for f in dataclasses.fields(cls)},
            keys_supplied,
        )

    @classmethod
    def from_config(cls, config_dict):
        """
        Attempt to instantiate an object of the class on which it is called,
        by means of the configuration settings passed in.

        First, we check that the keys supplied in the dictionary correspond
        to data fields on the current class.
        Then, the dictionary is processed using the :meth:`process_entries`
        method. The resulting dictionary is passed to the initialiser
        of the current class as a kwargs dict.

        :param config_dict:
            A dictionary containing configuration values.
        :return:
            An instance of the class on which it is called.
        :raises ConfigurationError:
            when an unexpected configuration key is encountered or left
            unfilled, or when there is a problem processing one of the config
            values.
        """
        if not isinstance(config_dict, dict):
            raise ConfigurationError(
                f"{cls.__name__} requires a dictionary to initialise."
            )
        cls.check_config_keys(set(config_dict.keys()))
        # in Python we need underscores
        config_dict = {
            key.replace('-', '_'): v for key, v in config_dict.items()
        }

        cls.process_entries(config_dict)

        cls._process_configurable_fields(config_dict)

        enforce_required_keys(
            cls.__name__,
            {f.name for f in dataclasses.fields(cls) if not _has_default(f)},
            config_dict,
        )
        try:
            # noinspection PyArgumentList
            return cls(**config_dict)
        except TypeError as e:  # pragma: nocover
            raise ConfigurationError("Failed to instantiate from config") from e


def check_config_keys(config_name, expected_keys, supplied_keys):
    # wrapper function to provide user-friendly errors
    #  (mainly intended for the CLI)
    # This does not check whether all required keys are present, that happens
    # later
    # TODO What about type checking?
    unexpected_keys = _check_subset(supplied_keys, expected_keys)
    if unexpected_keys:
        # this is easier to present to the user than a TypeError
        raise ConfigurationError(
            f"Unexpected {'key' if len(unexpected_keys) == 1 else 'keys'} "
            f"in configuration for {config_name}: "
            f"{', '.join(unexpected_keys)}."
        )


def _check_subset(expected_sub, expected_sup):
    # standardise on dashes for the yaml interface
    expected_sub = {key.replace('_', '-') for key in expected_sub}
    expected_sup = {key.replace('_', '-') for key in expected_sup}
    return expected_sub - expected_sup


def enforce_required_keys(config_name, required_keys, config_dict):
    missing_keys = _check_subset(required_keys, config_dict.keys())
    if missing_keys:
        # this is easier to present to the user than a TypeError
        raise ConfigurationError(
            f"Missing required {'key' if len(missing_keys) == 1 else 'keys'} "
            f"in configuration for {config_name}: "
            f"{', '.join(missing_keys)}."
        )


OID_REGEX = re.compile(r'\d(\.\d+)+')


def process_oid(
    asn1crypto_class: Type[ObjectIdentifier], id_string, param_name
):
    if not isinstance(id_string, str):
        raise ConfigurationError(
            f"Identifier '{repr(id_string)}' in '{param_name}' is not a string."
        )
    if OID_REGEX.fullmatch(id_string):
        # Attempt to translate OID strings to human-readable form
        # whenever possible
        return asn1crypto_class.map(id_string)
    else:
        try:
            # try to see if the usage string maps to an OID before
            # proceeding
            asn1crypto_class.unmap(id_string)
        except ValueError:
            raise ConfigurationError(
                f"'{id_string}' in '{param_name}' is not a valid "
                f"{asn1crypto_class.__name__}"
            )
        return id_string


def _ensure_strings(strings, param_name):
    err_msg = (
        f"'{param_name}' must be specified as a "
        "list of strings, or a string."
    )
    if isinstance(strings, str):
        strings = (strings,)
    elif not isinstance(strings, (list, tuple)):
        raise ConfigurationError(err_msg)

    return strings


def process_oids(asn1crypto_class: Type[ObjectIdentifier], strings, param_name):
    strings = _ensure_strings(strings, param_name)
    for usage_string in strings:
        yield process_oid(asn1crypto_class, usage_string, param_name)


def process_bit_string_flags(
    asn1crypto_class: Type[BitString], strings, param_name
):
    strings = _ensure_strings(strings, param_name)
    valid_values = asn1crypto_class._map.values()
    for flag_string in strings:
        if not isinstance(flag_string, str):
            raise ConfigurationError(
                f"Flag identifier '{repr(flag_string)}' is not a string."
            )
        elif flag_string not in valid_values:
            raise ConfigurationError(
                f"'{repr(flag_string)}' is not a valid "
                f"{asn1crypto_class.__name__} flag name."
            )
        # Use yield to keep the API consistent with process_oids, we don't
        # change anything
        yield flag_string



"""
Internal constants for the difference analysis sub-package.
"""

import re

__all__ = [
    'ACROFORM_EXEMPT_STRICT_COMPARISON',
    'ROOT_EXEMPT_STRICT_COMPARISON',
    'FORMFIELD_ALWAYS_MODIFIABLE',
    'VALUE_UPDATE_KEYS',
    'VRI_KEY_PATTERN',
]

# /Type: dictionary type (can always be added if correct)
# /Ff: Form field flags
FORMFIELD_ALWAYS_MODIFIABLE = frozenset({'/Ff', '/Type'})

# /AP: appearance dictionary
# /AS: current appearance state
# /V: field value
# /F: (widget) annotation flags
# /DA: default appearance
# /Q: quadding
VALUE_UPDATE_KEYS = FORMFIELD_ALWAYS_MODIFIABLE | frozenset(
    {
        '/AP',
        '/AS',
        '/V',
        '/F',
        '/DA',
        '/Q',
    }
)

VRI_KEY_PATTERN = re.compile('/[A-Z0-9]{40}')

ACROFORM_EXEMPT_STRICT_COMPARISON = frozenset(
    {
        '/Fields',
        '/DR',
        '/DA',
        '/Q',
        '/NeedAppearances',
    }
)

ROOT_EXEMPT_STRICT_COMPARISON = frozenset(
    {
        '/AcroForm',
        '/DSS',
        '/Extensions',
        '/Metadata',
        '/MarkInfo',
        '/Version',
    }
)



from io import BytesIO

import pytest
from pyhanko.pdf_utils import generic, text
from pyhanko.pdf_utils.font.opentype import GlyphAccumulator
from pyhanko.pdf_utils.generic import pdf_name
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.layout import BoxConstraints
from pyhanko.pdf_utils.reader import PdfFileReader

from .samples import MINIMAL, MINIMAL_XREF, TEST_DIR


@pytest.mark.parametrize(
    'with_border, natural_size', [[True, True], [False, False], [True, False]]
)
def test_simple_textbox_render(with_border, natural_size):
    tbs = text.TextBoxStyle(border_width=1 if with_border else 0)
    bc = None if natural_size else BoxConstraints(width=1600, height=900)

    textbox = text.TextBox(
        style=tbs, box=bc, writer=IncrementalPdfFileWriter(BytesIO(MINIMAL))
    )
    textbox.content = 'This is a textbox with some text.\nAnd multiple lines'
    xobj = textbox.as_form_xobject()
    x1, y1, x2, y2 = xobj['/BBox']
    assert '/F1' in textbox.resources.font

    if not natural_size:
        assert abs(x1 - x2) == 1600
        assert abs(y1 - y2) == 900


NOTO_SERIF_JP = f'{TEST_DIR}/data/fonts/NotoSerifJP-Regular.otf'


def test_embed_subset():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    with open(NOTO_SERIF_JP, 'rb') as ffile:
        ga = GlyphAccumulator(w, ffile, font_size=10)
    res = ga.shape('版')
    assert b'[<66eb>] TJ' in res.graphics_ops
    res = ga.shape('テスト')
    assert b'[<0637> 40 <062a0639>] TJ' in res.graphics_ops

    # check the 'ffi' ligature
    res = ga.shape('difficult')
    assert b'[<0045004ae9e200440056004d0055>] TJ' in res.graphics_ops

    w.write_in_place()
    font_ref = ga.as_resource()
    font = font_ref.get_object()
    df = font['/DescendantFonts'][0].get_object()
    font_file = df['/FontDescriptor']['/FontFile3']
    # assert no ToUnicode assignment for 'f'
    assert b'\n<0066>' not in font['/ToUnicode'].data
    # assert a ToUnicode assignment for the 'ffi' ligature
    assert b'\n<e9e2> <006600660069>' in font['/ToUnicode'].data
    assert len(font_file.data) < 4000


def test_actual_text_toggle():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    with open(NOTO_SERIF_JP, 'rb') as ffile:
        ga = GlyphAccumulator(w, ffile, font_size=10)

    res = ga.shape('difficult')
    assert b'[<0045004ae9e200440056004d0055>] TJ' in res.graphics_ops
    assert b'ActualText' in res.graphics_ops

    res = ga.shape('difficult', with_actual_text=False)
    assert b'[<0045004ae9e200440056004d0055>] TJ' in res.graphics_ops
    assert b'ActualText' not in res.graphics_ops


def test_write_embedded_string():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    with open(NOTO_SERIF_JP, 'rb') as ffile:
        ga = GlyphAccumulator(w, ffile, font_size=10)
    # shape the string, just to register the glyphs as used
    ga.shape('テスト')
    # ... but we're not going to use the result

    # hardcoded CIDs
    cid_hx = '0637062a0639'
    stream = generic.StreamObject(
        stream_data=f'BT /FEmb 18 Tf 0 100 Td <{cid_hx}> Tj ET'.encode('ascii')
    )
    stream_ref = w.add_object(stream)
    w.add_stream_to_page(
        0,
        stream_ref,
        resources=generic.DictionaryObject(
            {
                pdf_name('/Font'): generic.DictionaryObject(
                    {pdf_name('/FEmb'): ga.as_resource()}
                )
            }
        ),
    )
    out = BytesIO()
    w.write(out)
    out.seek(0)
    r = PdfFileReader(out)
    page_obj = r.root['/Pages']['/Kids'][0].get_object()
    conts = page_obj['/Contents']
    assert len(conts) == 2
    assert stream_ref.idnum in (c.idnum for c in conts)


def test_write_embedded_string_objstream():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_XREF))
    obj_stream = w.prepare_object_stream()
    with open(NOTO_SERIF_JP, 'rb') as ffile:
        ga = GlyphAccumulator(w, ffile, font_size=10, obj_stream=obj_stream)
    # shape the string, just to register the glyphs as used
    ga.shape('テスト')
    # ... but we're not going to use the result

    # hardcoded CIDs
    cid_hx = '0637062a0639'
    font_ref = ga.as_resource()
    stream = generic.StreamObject(
        stream_data=f'BT /FEmb 18 Tf 0 100 Td <{cid_hx}> Tj ET'.encode('ascii')
    )
    stream_ref = w.add_object(stream)
    w.add_stream_to_page(
        0,
        stream_ref,
        resources=generic.DictionaryObject(
            {
                pdf_name('/Font'): generic.DictionaryObject(
                    {pdf_name('/FEmb'): font_ref}
                )
            }
        ),
    )
    out = BytesIO()
    w.write(out)
    out.seek(0)
    r = PdfFileReader(out)
    page_obj = r.root['/Pages']['/Kids'][0].get_object()
    conts = page_obj['/Contents']
    assert len(conts) == 2
    assert stream_ref.idnum in (c.idnum for c in conts)
    xref_sections = r.xrefs._xref_sections
    last = xref_sections[len(xref_sections) - 1]
    assert font_ref.idnum in last.xref_data.xrefs_in_objstm
    out.seek(0)

    # attempt to grab the font from the object stream
    font_ref.pdf = r
    font = font_ref.get_object()
    assert font['/Type'] == pdf_name('/Font')


def test_opentype_with_langsys():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    with open(
        # subset rigged for exactly this test so we don't have to
        # ship the entire font file
        f"{TEST_DIR}/data/fonts/NotoSerifSubset.otf",
        'rb',
    ) as ffile:
        ga1 = GlyphAccumulator(
            w,
            ffile,
            font_size=10,
        )
        ga2 = GlyphAccumulator(
            w, ffile, font_size=10, ot_language_tag='ZHT ', ot_script_tag='hani'
        )
    shape_result1 = ga1.shape('骨')
    shape_result2 = ga2.shape('骨')
    assert b'<b04c>' in shape_result1.graphics_ops
    assert b'<b04e>' in shape_result2.graphics_ops


def test_opentype_langsys_tag_too_long():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    with open(
        f"{TEST_DIR}/data/fonts/NotoSerifSubset.otf",
        'rb',
    ) as ffile:
        with pytest.raises(ValueError, match='4 bytes long'):
            GlyphAccumulator(
                w,
                ffile,
                font_size=10,
                ot_language_tag='ZHT a',
                ot_script_tag='hani',
            )


def test_opentype_langsys_tag_not_ascii():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    with open(
        f"{TEST_DIR}/data/fonts/NotoSerifSubset.otf",
        'rb',
    ) as ffile:
        with pytest.raises(ValueError, match='encodable'):
            GlyphAccumulator(
                w,
                ffile,
                font_size=10,
                ot_language_tag='ZHTあ',
                ot_script_tag='hani',
            )


def test_opentype_invalid_writing_direction():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    with open(
        f"{TEST_DIR}/data/fonts/NotoSerifSubset.otf",
        'rb',
    ) as ffile:
        with pytest.raises(ValueError, match='direction must be one of'):
            GlyphAccumulator(w, ffile, font_size=10, writing_direction='foobar')



import binascii
import os
from io import BytesIO
from typing import Optional

import pytest
from asn1crypto import cms, x509
from certomancer.registry import ArchLabel, CertLabel, KeyLabel
from pyhanko.keys import load_cert_from_pemder
from pyhanko.pdf_utils import generic, misc, writer
from pyhanko.pdf_utils.crypt import (
    DEFAULT_CRYPT_FILTER,
    STD_CF,
    AuthStatus,
    CryptFilterConfiguration,
    IdentityCryptFilter,
    PdfKeyNotAvailableError,
    PubKeyAdbeSubFilter,
    PubKeyAESCryptFilter,
    PubKeyRC4CryptFilter,
    PubKeySecurityHandler,
    SecurityHandler,
    SecurityHandlerVersion,
    SerialisedCredential,
    SimpleEnvelopeKeyDecrypter,
    StandardAESCryptFilter,
    StandardRC4CryptFilter,
    StandardSecurityHandler,
    StandardSecuritySettingsRevision,
    build_crypt_filter,
    pubkey,
)
from pyhanko.pdf_utils.crypt.permissions import (
    PubKeyPermissions,
    StandardPermissions,
)
from pyhanko.pdf_utils.crypt.standard import StandardAESGCMCryptFilter
from pyhanko.pdf_utils.generic import pdf_name
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.reader import PdfFileReader

from .samples import (
    CERTOMANCER,
    MINIMAL,
    MINIMAL_AES256,
    MINIMAL_ONE_FIELD,
    MINIMAL_ONE_FIELD_AES256,
    PDF_DATA_DIR,
    PUBKEY_SELFSIGNED_DECRYPTER,
    PUBKEY_TEST_DECRYPTER,
    PUBKEY_TEST_DECRYPTER_OLD,
    TEST_DIR,
    TESTING_CA_DIR,
    VECTOR_IMAGE_PDF,
)

STD_PERMS = (
    ~StandardPermissions.ALLOW_MODIFICATION_GENERIC
    & ~StandardPermissions.ALLOW_ANNOTS_FORM_FILLING
)

PUBKEY_PERMS = (
    ~PubKeyPermissions.ALLOW_ENCRYPTION_CHANGE
    & ~PubKeyPermissions.ALLOW_MODIFICATION_GENERIC
    & ~PubKeyPermissions.ALLOW_ANNOTS_FORM_FILLING
)


def _produce_legacy_encrypted_file(rev, keylen_bytes, use_aes):
    r = PdfFileReader(BytesIO(VECTOR_IMAGE_PDF))
    w = writer.PdfFileWriter()
    sh = StandardSecurityHandler.build_from_pw_legacy(
        rev,
        w._document_id[0].original_bytes,
        "ownersecret",
        "usersecret",
        keylen_bytes=keylen_bytes,
        use_aes128=use_aes,
        perms=STD_PERMS,
    )
    w._assign_security_handler(sh)
    new_page_tree = w.import_object(
        r.root.raw_get('/Pages'),
    )
    w.root['/Pages'] = new_page_tree
    out = BytesIO()
    w.write(out)
    return out


@pytest.mark.parametrize(
    "use_owner_pass,rev,keylen_bytes,use_aes",
    [
        (True, StandardSecuritySettingsRevision.RC4_BASIC, 5, False),
        (False, StandardSecuritySettingsRevision.RC4_BASIC, 5, False),
        (True, StandardSecuritySettingsRevision.RC4_EXTENDED, 5, False),
        (False, StandardSecuritySettingsRevision.RC4_EXTENDED, 5, False),
        (True, StandardSecuritySettingsRevision.RC4_EXTENDED, 16, False),
        (False, StandardSecuritySettingsRevision.RC4_EXTENDED, 16, False),
        (True, StandardSecuritySettingsRevision.RC4_OR_AES128, 5, False),
        (False, StandardSecuritySettingsRevision.RC4_OR_AES128, 5, False),
        (True, StandardSecuritySettingsRevision.RC4_OR_AES128, 16, False),
        (False, StandardSecuritySettingsRevision.RC4_OR_AES128, 16, False),
        (True, StandardSecuritySettingsRevision.RC4_OR_AES128, 16, True),
        (False, StandardSecuritySettingsRevision.RC4_OR_AES128, 16, True),
    ],
)
def test_legacy_encryption(use_owner_pass, rev, keylen_bytes, use_aes):
    out = _produce_legacy_encrypted_file(rev, keylen_bytes, use_aes)
    r = PdfFileReader(out)
    result = r.decrypt("ownersecret" if use_owner_pass else "usersecret")
    if use_owner_pass:
        assert result.status == AuthStatus.OWNER
    else:
        assert result.status == AuthStatus.USER
    assert result.permission_flags.as_sint32() == -44
    page = r.root['/Pages']['/Kids'][0].get_object()
    assert r.trailer['/Encrypt']['/P'] == -44
    assert '/ExtGState' in page['/Resources']
    # just a piece of data I know occurs in the decoded content stream
    # of the (only) page in VECTOR_IMAGE_PDF
    assert b'0 1 0 rg /a0 gs' in page['/Contents'].data


@pytest.mark.parametrize("legacy", [True, False])
def test_wrong_password(legacy):
    w = writer.PdfFileWriter()
    ref = w.add_object(generic.TextStringObject("Blah blah"))
    if legacy:
        sh = StandardSecurityHandler.build_from_pw_legacy(
            StandardSecuritySettingsRevision.RC4_OR_AES128,
            w._document_id[0].original_bytes,
            "ownersecret",
            "usersecret",
            keylen_bytes=16,
            use_aes128=True,
        )
    else:
        sh = StandardSecurityHandler.build_from_pw("ownersecret", "usersecret")
    w.security_handler = sh
    w._encrypt = w.add_object(sh.as_pdf_object())
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    with pytest.raises(misc.PdfReadError):
        r.get_object(ref.reference)
    assert r.decrypt("thispasswordiswrong").status == AuthStatus.FAILED
    assert r.security_handler._auth_failed
    assert r.security_handler.get_string_filter()._auth_failed
    with pytest.raises(misc.PdfReadError):
        r.get_object(ref.reference)


def test_identity_crypt_filter_api():
    # confirm that the CryptFilter API of the identity filter doesn't do
    # anything unexpected, even though we typically don't invoke it explicitly.
    idf: IdentityCryptFilter = IdentityCryptFilter()
    idf._set_security_handler(None)
    assert not idf._auth_failed
    assert isinstance(idf.derive_shared_encryption_key(), bytes)
    assert isinstance(idf.derive_object_key(1, 2), bytes)
    assert isinstance(idf.method, generic.NameObject)
    assert isinstance(idf.keylen, int)
    assert idf.decrypt(None, b'abc') == b'abc'
    assert idf.encrypt(None, b'abc') == b'abc'

    # can't serialise /Identity
    with pytest.raises(misc.PdfError):
        idf.as_pdf_object()


@pytest.mark.parametrize(
    "use_alias, with_never_decrypt",
    [(True, False), (False, True), (False, False)],
)
def test_identity_crypt_filter(use_alias, with_never_decrypt):
    w = writer.PdfFileWriter()
    sh = StandardSecurityHandler.build_from_pw("secret")
    w.security_handler = sh
    idf: IdentityCryptFilter = IdentityCryptFilter()
    assert sh.crypt_filter_config[pdf_name("/Identity")] is idf
    if use_alias:
        sh.crypt_filter_config._crypt_filters[pdf_name("/IdentityAlias")] = idf
        assert sh.crypt_filter_config[pdf_name("/IdentityAlias")] is idf
    if use_alias:
        # identity filter can't be serialised, so this should throw an error
        with pytest.raises(misc.PdfError):
            w._assign_security_handler(sh)
        return
    else:
        w._assign_security_handler(sh)
    test_bytes = b'This is some test data that should remain unencrypted.'
    test_stream = generic.StreamObject(stream_data=test_bytes, handler=sh)
    test_stream.apply_filter(
        "/Crypt", params={pdf_name("/Name"): pdf_name("/Identity")}
    )
    ref = w.add_object(test_stream).reference
    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    r.decrypt("secret")
    the_stream = r.get_object(ref, never_decrypt=with_never_decrypt)
    assert the_stream.encoded_data == test_bytes
    assert the_stream.data == test_bytes


def _produce_pubkey_encrypted_file(
    version,
    keylen,
    use_aes,
    use_crypt_filters,
    policy: pubkey.RecipientEncryptionPolicy = pubkey.RecipientEncryptionPolicy(),
    cert: Optional[x509.Certificate] = None,
):
    if cert is None:
        cert = PUBKEY_TEST_DECRYPTER.cert
    r = PdfFileReader(BytesIO(VECTOR_IMAGE_PDF))
    w = writer.PdfFileWriter()

    sh = PubKeySecurityHandler.build_from_certs(
        [cert],
        keylen_bytes=keylen,
        version=version,
        use_aes=use_aes,
        use_crypt_filters=use_crypt_filters,
        perms=PUBKEY_PERMS,
        policy=policy,
        pdf_mac=False,
    )
    w._assign_security_handler(sh)
    new_page_tree = w.import_object(
        r.root.raw_get('/Pages'),
    )
    w.root['/Pages'] = new_page_tree
    out = BytesIO()
    w.write(out)
    return out


def _validate_pubkey_decryption(r, result):
    assert result.status == AuthStatus.USER
    assert result.permission_flags == PUBKEY_PERMS
    page = r.root['/Pages']['/Kids'][0].get_object()
    assert '/ExtGState' in page['/Resources']
    # just a piece of data I know occurs in the decoded content stream
    # of the (only) page in VECTOR_IMAGE_PDF
    assert b'0 1 0 rg /a0 gs' in page['/Contents'].data


def _extract_ktri_kea(r: PdfFileReader):
    recp_bytes = r.encrypt_dict['/CF']['/DefaultCryptFilter']['/Recipients'][0]
    content_info = cms.ContentInfo.load(recp_bytes)
    ktri = content_info['content']['recipient_infos'][0].chosen
    return ktri['key_encryption_algorithm']['algorithm'].native


@pytest.mark.parametrize(
    "version, keylen, use_aes, use_crypt_filters",
    [
        (SecurityHandlerVersion.AES256, 32, True, True),
        (SecurityHandlerVersion.RC4_OR_AES128, 16, True, True),
        (SecurityHandlerVersion.RC4_OR_AES128, 16, False, True),
        (SecurityHandlerVersion.RC4_OR_AES128, 5, False, True),
        (SecurityHandlerVersion.RC4_40, 5, False, True),
        (SecurityHandlerVersion.RC4_40, 5, False, False),
        (SecurityHandlerVersion.RC4_LONGER_KEYS, 5, False, True),
        (SecurityHandlerVersion.RC4_LONGER_KEYS, 5, False, False),
        (SecurityHandlerVersion.RC4_LONGER_KEYS, 16, False, True),
        (SecurityHandlerVersion.RC4_LONGER_KEYS, 16, False, False),
    ],
)
def test_pubkey_encryption(version, keylen, use_aes, use_crypt_filters):
    out = _produce_pubkey_encrypted_file(
        version, keylen, use_aes, use_crypt_filters
    )
    r = PdfFileReader(out)
    if version == SecurityHandlerVersion.AES256:
        assert r.input_version == (2, 0)
    result = r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)
    _validate_pubkey_decryption(r, result)
    if version == SecurityHandlerVersion.AES256:
        kea = _extract_ktri_kea(r)
        assert kea == 'rsaes_pkcs1v15'


@pytest.mark.parametrize(
    'lbl',
    [
        'decrypter1',
        'decrypter2',
        'decrypter3',
        'decrypter-x25519',
        'decrypter-x448',
    ],
)
def test_ecdh_encryption(lbl):
    pki_arch = CERTOMANCER.get_pki_arch(
        ArchLabel('ecc-testing-ca-with-decrypters')
    )
    cert = pki_arch.get_cert(CertLabel(lbl))
    out = _produce_pubkey_encrypted_file(
        version=SecurityHandlerVersion.AES256,
        keylen=32,
        use_aes=True,
        use_crypt_filters=True,
        cert=cert,
    )
    r = PdfFileReader(out)
    result = r.decrypt_pubkey(
        SimpleEnvelopeKeyDecrypter(
            cert, pki_arch.key_set.get_private_key(KeyLabel(lbl))
        )
    )
    _validate_pubkey_decryption(r, result)


def test_oaep_encryption():
    out = _produce_pubkey_encrypted_file(
        version=SecurityHandlerVersion.AES256,
        keylen=32,
        use_aes=True,
        use_crypt_filters=True,
        policy=pubkey.RecipientEncryptionPolicy(prefer_oaep=True),
    )
    r = PdfFileReader(out)
    result = r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)
    _validate_pubkey_decryption(r, result)
    kea = _extract_ktri_kea(r)
    assert kea == 'rsaes_oaep'


def test_ecdh_decryption_smoke():
    pki_arch = CERTOMANCER.get_pki_arch(
        ArchLabel('ecc-testing-ca-with-decrypters')
    )
    cert = pki_arch.get_cert(CertLabel('decrypter3'))
    decrypter = SimpleEnvelopeKeyDecrypter(
        cert, pki_arch.key_set.get_private_key(KeyLabel('decrypter3'))
    )
    with open(f'{PDF_DATA_DIR}/pubkey-ecc-test.pdf', 'rb') as inf:
        r = PdfFileReader(inf)
        result = r.decrypt_pubkey(decrypter)
        assert result.status == AuthStatus.USER
        assert result.permission_flags == PubKeyPermissions.allow_everything()


def test_ecdh_decryption_wrong_key_type():
    pki_arch = CERTOMANCER.get_pki_arch(
        ArchLabel('ecc-testing-ca-with-decrypters')
    )
    cert = pki_arch.get_cert(CertLabel('decrypter3'))
    with open(f'{PDF_DATA_DIR}/pubkey-ecc-test.pdf', 'rb') as inf:
        r = PdfFileReader(inf)
        decrypter = SimpleEnvelopeKeyDecrypter(
            cert=cert, private_key=PUBKEY_TEST_DECRYPTER.private_key
        )
        with pytest.raises(pubkey.InappropriateCredentialError):
            r.decrypt_pubkey(decrypter)


def test_ecdh_decryption_wrong_curve():
    pki_arch = CERTOMANCER.get_pki_arch(
        ArchLabel('ecc-testing-ca-with-decrypters')
    )
    cert = pki_arch.get_cert(CertLabel('decrypter3'))
    with open(f'{PDF_DATA_DIR}/pubkey-ecc-test.pdf', 'rb') as inf:
        r = PdfFileReader(inf)
        decrypter = SimpleEnvelopeKeyDecrypter(
            cert=cert,
            private_key=pki_arch.key_set.get_private_key(
                KeyLabel('decrypter1')
            ),
        )
        with pytest.raises(misc.PdfReadError, match='Failed to decrypt'):
            r.decrypt_pubkey(decrypter)


def test_rsa_decryption_wrong_key_type():
    pki_arch = CERTOMANCER.get_pki_arch(
        ArchLabel('ecc-testing-ca-with-decrypters')
    )
    key = pki_arch.key_set.get_private_key(KeyLabel('decrypter1'))
    with open(f'{PDF_DATA_DIR}/minimal-pubkey-rc4-envelope.pdf', 'rb') as inf:
        r = PdfFileReader(inf)
        decrypter = SimpleEnvelopeKeyDecrypter(
            cert=PUBKEY_TEST_DECRYPTER_OLD.cert, private_key=key
        )
        with pytest.raises(pubkey.InappropriateCredentialError):
            r.decrypt_pubkey(decrypter)


@pytest.mark.parametrize(
    'lbl', ['decrypter1', 'decrypter-x25519', 'decrypter-x448']
)
def test_invalid_originator_key(monkeypatch, lbl):
    pki_arch = CERTOMANCER.get_pki_arch(
        ArchLabel('ecc-testing-ca-with-decrypters')
    )
    cert = pki_arch.get_cert(CertLabel(lbl))
    priv_key = pki_arch.key_set.get_private_key(KeyLabel(lbl))

    def _format_kari(
        rid,
        originator_key,
        algo,
        ukm,
        encrypted_data,
    ):
        return cms.RecipientInfo(
            name='kari',
            value=cms.KeyAgreeRecipientInfo(
                {
                    'version': 3,
                    'originator': cms.OriginatorIdentifierOrKey(
                        name='originator_key',
                        value=PUBKEY_TEST_DECRYPTER.cert.public_key,
                    ),
                    'ukm': ukm,
                    'key_encryption_algorithm': algo,
                    'recipient_encrypted_keys': [
                        cms.RecipientEncryptedKey(
                            {'rid': rid, 'encrypted_key': encrypted_data}
                        )
                    ],
                }
            ),
        )

    monkeypatch.setattr(pubkey, '_format_kari', _format_kari)
    out = _produce_pubkey_encrypted_file(
        version=SecurityHandlerVersion.AES256,
        keylen=32,
        use_aes=True,
        use_crypt_filters=True,
        cert=cert,
    )

    r = PdfFileReader(out)
    decrypter = SimpleEnvelopeKeyDecrypter(
        cert=cert,
        private_key=priv_key,
    )
    with pytest.raises(misc.PdfReadError):
        r.decrypt_pubkey(decrypter)


def test_key_encipherment_requirement():
    with pytest.raises(misc.PdfWriteError):
        PubKeySecurityHandler.build_from_certs(
            [PUBKEY_SELFSIGNED_DECRYPTER.cert],
            keylen_bytes=32,
            version=SecurityHandlerVersion.AES256,
            use_aes=True,
            use_crypt_filters=True,
            perms=PUBKEY_PERMS,
            pdf_mac=False,
        )


@pytest.mark.parametrize(
    "version, keylen, use_aes, use_crypt_filters",
    [
        (SecurityHandlerVersion.AES256, 32, True, True),
        (SecurityHandlerVersion.RC4_OR_AES128, 16, True, True),
        (SecurityHandlerVersion.RC4_OR_AES128, 16, False, True),
        (SecurityHandlerVersion.RC4_OR_AES128, 5, False, True),
        (SecurityHandlerVersion.RC4_40, 5, False, True),
        (SecurityHandlerVersion.RC4_40, 5, False, False),
        (SecurityHandlerVersion.RC4_LONGER_KEYS, 5, False, True),
        (SecurityHandlerVersion.RC4_LONGER_KEYS, 5, False, False),
        (SecurityHandlerVersion.RC4_LONGER_KEYS, 16, False, True),
        (SecurityHandlerVersion.RC4_LONGER_KEYS, 16, False, False),
    ],
)
def test_key_encipherment_requirement_override(
    version, keylen, use_aes, use_crypt_filters
):
    r = PdfFileReader(BytesIO(VECTOR_IMAGE_PDF))
    w = writer.PdfFileWriter()

    sh = PubKeySecurityHandler.build_from_certs(
        [PUBKEY_SELFSIGNED_DECRYPTER.cert],
        keylen_bytes=keylen,
        version=version,
        use_aes=use_aes,
        use_crypt_filters=use_crypt_filters,
        perms=PUBKEY_PERMS,
        policy=pubkey.RecipientEncryptionPolicy(ignore_key_usage=True),
        pdf_mac=False,
    )
    w._assign_security_handler(sh)
    new_page_tree = w.import_object(
        r.root.raw_get('/Pages'),
    )
    w.root['/Pages'] = new_page_tree
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    result = r.decrypt_pubkey(PUBKEY_SELFSIGNED_DECRYPTER)
    _validate_pubkey_decryption(r, result)


def test_pubkey_alternative_filter():
    w = writer.PdfFileWriter()

    w.encrypt_pubkey([PUBKEY_TEST_DECRYPTER.cert])
    # subfilter should be picked up
    w._encrypt.get_object()['/Filter'] = pdf_name('/FooBar')
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    assert isinstance(r.security_handler, PubKeySecurityHandler)


@pytest.mark.parametrize('delete_subfilter', [True, False])
def test_pubkey_unsupported_filter(delete_subfilter):
    w = writer.PdfFileWriter()

    w.encrypt_pubkey([PUBKEY_TEST_DECRYPTER.cert])
    encrypt = w._encrypt.get_object()
    encrypt['/Filter'] = pdf_name('/FooBar')
    if delete_subfilter:
        del encrypt['/SubFilter']
    else:
        encrypt['/SubFilter'] = pdf_name('/baz.quux')
    out = BytesIO()
    w.write(out)
    with pytest.raises(misc.PdfReadError):
        # noinspection PyStatementEffect
        PdfFileReader(out).root['/Pages']['/Kids'][0]['/Content'].data


def test_pubkey_encryption_block_cfs_s4():
    w = writer.PdfFileWriter()

    w.encrypt_pubkey([PUBKEY_TEST_DECRYPTER.cert])
    encrypt = w._encrypt.get_object()
    encrypt['/SubFilter'] = pdf_name('/adbe.pkcs7.s4')
    out = BytesIO()
    w.write(out)
    with pytest.raises(misc.PdfReadError):
        # noinspection PyStatementEffect
        PdfFileReader(out).root['/Pages']['/Kids'][0]['/Content'].data


def test_pubkey_encryption_s5_requires_cfs():
    w = writer.PdfFileWriter()

    sh = PubKeySecurityHandler.build_from_certs([PUBKEY_TEST_DECRYPTER.cert])
    w._assign_security_handler(sh)
    encrypt = w._encrypt.get_object()
    del encrypt['/CF']
    out = BytesIO()
    w.write(out)
    with pytest.raises(misc.PdfReadError):
        # noinspection PyStatementEffect
        PdfFileReader(out).root['/Pages']['/Kids'][0]['/Content'].data


def test_pubkey_encryption_dict_errors():
    sh = PubKeySecurityHandler.build_from_certs([PUBKEY_TEST_DECRYPTER.cert])

    encrypt = generic.DictionaryObject(sh.as_pdf_object())
    encrypt['/SubFilter'] = pdf_name('/asdflakdsjf')
    with pytest.raises(misc.PdfReadError):
        PubKeySecurityHandler.build(encrypt)

    encrypt = generic.DictionaryObject(sh.as_pdf_object())
    encrypt['/Length'] = generic.NumberObject(13)
    with pytest.raises(misc.PdfError):
        PubKeySecurityHandler.build(encrypt)

    encrypt = generic.DictionaryObject(sh.as_pdf_object())
    del encrypt['/CF']['/DefaultCryptFilter']['/CFM']
    with pytest.raises(misc.PdfReadError):
        PubKeySecurityHandler.build(encrypt)

    encrypt = generic.DictionaryObject(sh.as_pdf_object())
    del encrypt['/CF']['/DefaultCryptFilter']['/Recipients']
    with pytest.raises(misc.PdfReadError):
        PubKeySecurityHandler.build(encrypt)

    encrypt = generic.DictionaryObject(sh.as_pdf_object())
    encrypt['/CF']['/DefaultCryptFilter']['/CFM'] = pdf_name('/None')
    with pytest.raises(misc.PdfReadError):
        PubKeySecurityHandler.build(encrypt)


@pytest.mark.parametrize(
    'with_hex_filter, main_unencrypted',
    [(True, False), (True, True), (False, True), (False, False)],
)
def test_custom_crypt_filter(with_hex_filter, main_unencrypted):
    w = writer.PdfFileWriter()
    custom = pdf_name('/Custom')
    crypt_filters = {
        custom: StandardRC4CryptFilter(keylen=16),
    }
    if main_unencrypted:
        # streams/strings are unencrypted by default
        cfc = CryptFilterConfiguration(crypt_filters=crypt_filters)
        assert len(cfc.filters()) == 1
    else:
        crypt_filters[STD_CF] = StandardAESCryptFilter(keylen=16)
        cfc = CryptFilterConfiguration(
            crypt_filters=crypt_filters,
            default_string_filter=STD_CF,
            default_stream_filter=STD_CF,
        )
        assert len(cfc.filters()) == 2
    sh = StandardSecurityHandler.build_from_pw_legacy(
        rev=StandardSecuritySettingsRevision.RC4_OR_AES128,
        id1=w.document_id[0],
        desired_user_pass="usersecret",
        desired_owner_pass="ownersecret",
        keylen_bytes=16,
        crypt_filter_config=cfc,
    )
    w._assign_security_handler(sh)
    test_data = b'This is test data!'
    dummy_stream = generic.StreamObject(stream_data=test_data)
    dummy_stream.add_crypt_filter(name=custom, handler=sh)
    ref = w.add_object(dummy_stream)
    dummy_stream2 = generic.StreamObject(stream_data=test_data)
    ref2 = w.add_object(dummy_stream2)

    if with_hex_filter:
        dummy_stream.apply_filter(pdf_name('/AHx'))
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    r.decrypt("ownersecret")
    obj: generic.StreamObject = r.get_object(ref.reference)
    assert obj.data == test_data
    if with_hex_filter:
        cf_dict = obj['/DecodeParms'][1]
    else:
        cf_dict = obj['/DecodeParms']

    assert cf_dict['/Name'] == pdf_name('/Custom')

    obj2: generic.DecryptedObjectProxy = r.get_object(
        ref2.reference, transparent_decrypt=False
    )
    raw = obj2.raw_object
    assert isinstance(raw, generic.StreamObject)
    if main_unencrypted:
        assert raw.encoded_data == test_data
    else:
        assert raw.encoded_data != test_data


@pytest.mark.parametrize(
    'with_hex_filter, main_unencrypted',
    [(True, False), (True, True), (False, True), (False, False)],
)
def test_custom_pubkey_crypt_filter(with_hex_filter, main_unencrypted):
    w = writer.PdfFileWriter()
    custom = pdf_name('/Custom')
    crypt_filters = {
        custom: PubKeyRC4CryptFilter(keylen=16),
    }
    if main_unencrypted:
        # streams/strings are unencrypted by default
        cfc = CryptFilterConfiguration(crypt_filters=crypt_filters)
    else:
        crypt_filters[DEFAULT_CRYPT_FILTER] = PubKeyAESCryptFilter(
            keylen=16, acts_as_default=True
        )
        cfc = CryptFilterConfiguration(
            crypt_filters=crypt_filters,
            default_string_filter=DEFAULT_CRYPT_FILTER,
            default_stream_filter=DEFAULT_CRYPT_FILTER,
        )
    sh = PubKeySecurityHandler(
        version=SecurityHandlerVersion.RC4_OR_AES128,
        pubkey_handler_subfilter=PubKeyAdbeSubFilter.S5,
        legacy_keylen=16,
        crypt_filter_config=cfc,
    )

    # if main_unencrypted, these should be no-ops
    sh.add_recipients([PUBKEY_TEST_DECRYPTER.cert])
    # (this is always pointless, but it should be allowed)
    sh.add_recipients([PUBKEY_TEST_DECRYPTER.cert])

    crypt_filters[custom].add_recipients(
        [PUBKEY_TEST_DECRYPTER.cert], policy=pubkey.RecipientEncryptionPolicy()
    )
    w._assign_security_handler(sh)

    encrypt_dict = w._encrypt.get_object()
    cfs = encrypt_dict['/CF']
    # no /Recipients in S5 mode
    assert '/Recipients' not in encrypt_dict
    assert isinstance(cfs[custom]['/Recipients'], generic.ByteStringObject)
    if main_unencrypted:
        assert DEFAULT_CRYPT_FILTER not in cfs
    else:
        default_rcpts = cfs[DEFAULT_CRYPT_FILTER]['/Recipients']
        assert isinstance(default_rcpts, generic.ArrayObject)
        assert len(default_rcpts) == 2

    # custom crypt filters can only have one set of recipients
    with pytest.raises(misc.PdfError):
        crypt_filters[custom].add_recipients(
            [PUBKEY_TEST_DECRYPTER.cert],
            policy=pubkey.RecipientEncryptionPolicy(),
        )

    test_data = b'This is test data!'
    dummy_stream = generic.StreamObject(stream_data=test_data)
    dummy_stream.add_crypt_filter(name=custom, handler=sh)
    ref = w.add_object(dummy_stream)
    dummy_stream2 = generic.StreamObject(stream_data=test_data)
    ref2 = w.add_object(dummy_stream2)

    if with_hex_filter:
        dummy_stream.apply_filter(pdf_name('/AHx'))
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)

    # the custom test filter shouldn't have been decrypted yet
    # so attempting to decode the stream should cause the crypt filter
    # to throw an error
    obj: generic.StreamObject = r.get_object(ref.reference)
    with pytest.raises(misc.PdfError):
        # noinspection PyStatementEffect
        obj.data

    r.security_handler.crypt_filter_config[custom].authenticate(
        PUBKEY_TEST_DECRYPTER
    )
    assert obj.data == test_data
    if with_hex_filter:
        cf_dict = obj['/DecodeParms'][1]
    else:
        cf_dict = obj['/DecodeParms']

    assert cf_dict['/Name'] == pdf_name('/Custom')

    obj2: generic.DecryptedObjectProxy = r.get_object(
        ref2.reference, transparent_decrypt=False
    )
    raw = obj2.raw_object
    assert isinstance(raw, generic.StreamObject)
    if main_unencrypted:
        assert raw.encoded_data == test_data
    else:
        assert raw.encoded_data != test_data


def test_custom_crypt_filter_errors():
    w = writer.PdfFileWriter()
    custom = pdf_name('/Custom')
    crypt_filters = {
        custom: StandardRC4CryptFilter(keylen=16),
        STD_CF: StandardAESCryptFilter(keylen=16),
    }
    cfc = CryptFilterConfiguration(
        crypt_filters=crypt_filters,
        default_string_filter=STD_CF,
        default_stream_filter=STD_CF,
    )
    sh = StandardSecurityHandler.build_from_pw_legacy(
        rev=StandardSecuritySettingsRevision.RC4_OR_AES128,
        id1=w.document_id[0],
        desired_user_pass="usersecret",
        desired_owner_pass="ownersecret",
        keylen_bytes=16,
        crypt_filter_config=cfc,
    )
    w._assign_security_handler(sh)
    test_data = b'This is test data!'
    dummy_stream = generic.StreamObject(stream_data=test_data)
    with pytest.raises(misc.PdfStreamError):
        dummy_stream.add_crypt_filter(name='/Idontexist', handler=sh)

    # no handler
    dummy_stream.add_crypt_filter(name=custom)
    dummy_stream._handler = None
    w.add_object(dummy_stream)

    out = BytesIO()
    with pytest.raises(misc.PdfStreamError):
        w.write(out)


@pytest.mark.parametrize('pdf_mac', [True, False])
def test_continue_encrypted_file_without_auth(pdf_mac):
    w = writer.PdfFileWriter()
    w.root["/Test"] = generic.TextStringObject("Blah blah")
    w.encrypt("ownersecret", "usersecret", pdf_mac=pdf_mac)
    out = BytesIO()
    w.write(out)
    incr_w = IncrementalPdfFileWriter(out)
    incr_w.root["/Test"] = generic.TextStringObject("Bluh bluh")
    incr_w.update_root()
    with pytest.raises(PdfKeyNotAvailableError):
        incr_w.write_in_place()


def test_continue_encrypted_file_without_auth_disable_meta():
    w = writer.PdfFileWriter()
    w.root["/Test"] = generic.TextStringObject("Blah blah")
    w.encrypt("ownersecret", "usersecret")
    out = BytesIO()
    w.write(out)
    incr_w = IncrementalPdfFileWriter(out)
    incr_w._update_meta = lambda: None
    incr_w.root["/Test"] = generic.TextStringObject("Bluh bluh")
    incr_w.update_root()
    with pytest.raises(PdfKeyNotAvailableError):
        incr_w.write_in_place()


def test_continue_encrypted_file_without_auth_disable_meta_and_mac():
    w = writer.PdfFileWriter()
    w.root["/Test"] = generic.TextStringObject("Blah blah")
    w.encrypt("ownersecret", "usersecret", pdf_mac=False)
    out = BytesIO()
    w.write(out)
    incr_w = IncrementalPdfFileWriter(out)
    incr_w._update_meta = lambda: None
    incr_w.root["/Test"] = generic.TextStringObject("Bluh bluh")
    incr_w.update_root()
    with pytest.raises(misc.PdfWriteError, match="Cannot update"):
        incr_w.write_in_place()


def test_continue_encrypted_file_from_reader():
    w = writer.PdfFileWriter()
    w.root["/Test"] = generic.TextStringObject("Blah blah")
    w.encrypt("ownersecret", "usersecret")
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    # first decrypt, then extend
    r.decrypt("usersecret")
    incr_w = IncrementalPdfFileWriter.from_reader(r)
    incr_w.root["/Test"] = generic.TextStringObject("Bluh bluh")
    incr_w.update_root()
    incr_w.write_in_place()

    r = PdfFileReader(out)
    r.decrypt("usersecret")
    assert r.root['/Test'] == generic.TextStringObject("Bluh bluh")


def test_aes256_perm_read():
    r = PdfFileReader(BytesIO(MINIMAL_ONE_FIELD_AES256))
    result = r.decrypt("ownersecret")
    assert result.status == AuthStatus.OWNER
    assert result.permission_flags == StandardPermissions.allow_everything()
    r = PdfFileReader(BytesIO(MINIMAL_ONE_FIELD_AES256))
    result = r.decrypt("usersecret")
    assert result.status == AuthStatus.USER
    assert result.permission_flags == StandardPermissions.allow_everything()

    assert r.trailer['/Encrypt']['/P'] == -4


def test_copy_encrypted_file():
    r = PdfFileReader(BytesIO(MINIMAL_ONE_FIELD_AES256))
    r.decrypt("ownersecret")
    w = writer.copy_into_new_writer(r)
    old_root_ref = w.root_ref
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    assert r.root_ref == old_root_ref
    assert len(r.root['/AcroForm']['/Fields']) == 1
    assert len(r.root['/Pages']['/Kids']) == 1


def test_copy_to_encrypted_file():
    r = PdfFileReader(BytesIO(MINIMAL_ONE_FIELD))
    w = writer.copy_into_new_writer(r)
    old_root_ref = w.root_ref
    w.encrypt("ownersecret", "usersecret")
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    result = r.decrypt("ownersecret")
    assert result.status == AuthStatus.OWNER
    assert r.root_ref == old_root_ref
    assert len(r.root['/AcroForm']['/Fields']) == 1
    assert len(r.root['/Pages']['/Kids']) == 1


def test_correctly_align_perms():
    r = PdfFileReader(BytesIO(MINIMAL_ONE_FIELD))
    w = writer.copy_into_new_writer(r)
    perms = ~StandardPermissions.allow_everything()
    w.encrypt("ownersecret", "usersecret", perms=perms)
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    result = r.decrypt("usersecret")
    assert result.status == AuthStatus.USER
    assert result.permission_flags == perms


def test_default_no_gcm():
    # document the fact that we don't yet apply ISO/TS 32003 by default
    r = PdfFileReader(BytesIO(MINIMAL))
    w = writer.copy_into_new_writer(r)
    w.encrypt("ownersecret", "usersecret")
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    assert r.decrypt("usersecret").status == AuthStatus.USER
    assert isinstance(
        r.security_handler.get_stream_filter(), StandardAESCryptFilter
    )


def test_gcm_via_encrypt_call():
    r = PdfFileReader(BytesIO(MINIMAL))
    w = writer.copy_into_new_writer(r)
    w.encrypt("ownersecret", "usersecret", use_gcm=True)
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    assert r.decrypt("usersecret").status == AuthStatus.USER
    assert isinstance(
        r.security_handler.get_stream_filter(), StandardAESGCMCryptFilter
    )


def test_empty_user_pass():
    r = PdfFileReader(BytesIO(MINIMAL_ONE_FIELD))
    w = writer.copy_into_new_writer(r)
    old_root_ref = w.root_ref
    w.encrypt('ownersecret', '')
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    result = r.decrypt('')
    assert result.status == AuthStatus.USER
    assert r.root_ref == old_root_ref
    assert len(r.root['/AcroForm']['/Fields']) == 1
    assert len(r.root['/Pages']['/Kids']) == 1
    assert r.root['/AcroForm']['/Fields'][0]['/T'] == 'Sig1'


def test_load_pkcs12():
    sedk = SimpleEnvelopeKeyDecrypter.load_pkcs12(
        f"{TEST_DIR}/data/crypto/selfsigned.pfx", b'exportsecret'
    )
    assert sedk.cert.subject == PUBKEY_SELFSIGNED_DECRYPTER.cert.subject


def test_pubkey_wrong_content_type():
    r = PdfFileReader(BytesIO(VECTOR_IMAGE_PDF))
    w = writer.PdfFileWriter()

    sh = PubKeySecurityHandler.build_from_certs(
        [PUBKEY_TEST_DECRYPTER.cert],
        version=SecurityHandlerVersion.RC4_40,
        keylen_bytes=16,
        use_aes=True,
        use_crypt_filters=False,
    )
    w.security_handler = sh
    enc_dict = sh.as_pdf_object()
    from asn1crypto import cms

    cms_bytes = cms.ContentInfo(
        {
            'content_type': cms.ContentType('data'),
            'content': cms.OctetString(b"\xde\xad\xbe\xef"),
        }
    ).dump()
    enc_dict['/Recipients'][0] = generic.ByteStringObject(cms_bytes)
    w._encrypt = w.add_object(enc_dict)
    new_page_tree = w.import_object(
        r.root.raw_get('/Pages'),
    )
    w.root['/Pages'] = new_page_tree
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    with pytest.raises(misc.PdfReadError, match="must be enveloped"):
        r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)


def test_pubkey_wrong_cert():
    r = PdfFileReader(BytesIO(VECTOR_IMAGE_PDF))
    w = writer.PdfFileWriter()

    recpt_cert = load_cert_from_pemder(
        TESTING_CA_DIR + '/interm/decrypter2.cert.pem'
    )
    test_data = b'This is test data!'
    dummy_stream = generic.StreamObject(stream_data=test_data)
    ref = w.add_object(dummy_stream)
    w.encrypt_pubkey([recpt_cert])
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    result = r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)
    assert result.status == AuthStatus.FAILED

    with pytest.raises(misc.PdfError):
        r.get_object(ref.reference)


def test_crypt_filter_build_failures():
    cfdict = generic.DictionaryObject()
    assert build_crypt_filter({}, cfdict, False) is None
    cfdict['/CFM'] = generic.NameObject('/None')
    assert build_crypt_filter({}, cfdict, False) is None

    with pytest.raises(NotImplementedError):
        cfdict['/CFM'] = generic.NameObject('/NoSuchCF')
        build_crypt_filter({}, cfdict, False)


@pytest.mark.parametrize('on_subclass', [True, False])
def test_custom_crypt_filter_type(on_subclass):
    w = writer.PdfFileWriter()
    custom_cf_type = pdf_name('/CustomCFType')

    class CustomCFClass(StandardRC4CryptFilter):
        def __init__(self):
            super().__init__(keylen=16)

        method = custom_cf_type

    if on_subclass:

        class NewStandardSecurityHandler(StandardSecurityHandler):
            pass

        sh_class = NewStandardSecurityHandler
        assert (
            sh_class._known_crypt_filters
            is not StandardSecurityHandler._known_crypt_filters
        )
        assert '/V2' in sh_class._known_crypt_filters
        SecurityHandler.register(sh_class)
    else:
        sh_class = StandardSecurityHandler

    sh_class.register_crypt_filter(
        custom_cf_type,
        lambda _, __: CustomCFClass(),
    )
    cfc = CryptFilterConfiguration(
        crypt_filters={STD_CF: CustomCFClass()},
        default_string_filter=STD_CF,
        default_stream_filter=STD_CF,
    )
    sh = sh_class.build_from_pw_legacy(
        rev=StandardSecuritySettingsRevision.RC4_OR_AES128,
        id1=w.document_id[0],
        desired_user_pass="usersecret",
        desired_owner_pass="ownersecret",
        keylen_bytes=16,
        crypt_filter_config=cfc,
    )
    assert isinstance(sh, sh_class)
    w._assign_security_handler(sh)
    test_data = b'This is test data!'
    dummy_stream = generic.StreamObject(stream_data=test_data)
    ref = w.add_object(dummy_stream)

    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    r.decrypt("ownersecret")

    cfc = r.security_handler.crypt_filter_config
    assert cfc.stream_filter_name == cfc.string_filter_name
    obj: generic.StreamObject = r.get_object(ref.reference)
    assert obj.data == test_data

    obj: generic.DecryptedObjectProxy = r.get_object(
        ref.reference, transparent_decrypt=False
    )
    assert isinstance(obj.raw_object, generic.StreamObject)
    assert obj.raw_object.encoded_data != test_data

    # restore security handler registry state
    del sh_class._known_crypt_filters[custom_cf_type]
    if on_subclass:
        SecurityHandler.register(StandardSecurityHandler)


def test_security_handler_version_deser():
    assert (
        SecurityHandlerVersion.from_number(5) == SecurityHandlerVersion.AES256
    )
    assert SecurityHandlerVersion.from_number(0) == SecurityHandlerVersion.OTHER
    assert (
        SecurityHandlerVersion.from_number(None) == SecurityHandlerVersion.OTHER
    )

    assert (
        StandardSecuritySettingsRevision.from_number(6)
        == StandardSecuritySettingsRevision.AES256
    )
    assert (
        StandardSecuritySettingsRevision.from_number(0)
        == StandardSecuritySettingsRevision.OTHER
    )


def test_key_len():
    with pytest.raises(misc.PdfError):
        SecurityHandlerVersion.RC4_OR_AES128.check_key_length(20)
    assert SecurityHandlerVersion.RC4_OR_AES128.check_key_length(6) == 6
    assert SecurityHandlerVersion.AES256.check_key_length(6) == 32
    assert SecurityHandlerVersion.RC4_40.check_key_length(32) == 5
    assert SecurityHandlerVersion.RC4_LONGER_KEYS.check_key_length(16) == 16


@pytest.mark.parametrize('pw', ['usersecret', 'ownersecret'])
def test_ser_deser_credential_standard_sh(pw):
    r = PdfFileReader(BytesIO(MINIMAL_AES256))
    r.decrypt(pw)
    cred = r.security_handler.extract_credential()
    assert cred['pwd_bytes'].native == pw.encode('utf8')
    cred_data = cred.serialise()

    r = PdfFileReader(BytesIO(MINIMAL_AES256))
    result = r.security_handler.authenticate(cred_data)
    exp_status = AuthStatus.USER if pw.startswith('user') else AuthStatus.OWNER
    assert result.status == exp_status


def test_ser_deser_credential_standard_sh_extract_from_builder():
    sh = StandardSecurityHandler.build_from_pw("ownersecret", "usersecret")
    cred = sh.extract_credential()
    assert cred['pwd_bytes'].native == b'ownersecret'
    assert cred['id1'].native is None


def test_ser_deser_credential_wrong_pw():
    r = PdfFileReader(BytesIO(MINIMAL_AES256))
    r.decrypt("ownersecret")
    cred = r.security_handler.extract_credential()
    cred['pwd_bytes'] = b'This is the wrong password'
    cred_data = cred.serialise()

    r = PdfFileReader(BytesIO(MINIMAL_AES256))
    result = r.security_handler.authenticate(cred_data)
    assert result.status == AuthStatus.FAILED


def test_ser_deser_credential_standard_corrupted():
    r = PdfFileReader(BytesIO(MINIMAL_AES256))
    r.decrypt("ownersecret")
    cred = r.security_handler.extract_credential()
    cred_data = SerialisedCredential(
        credential_type=cred.serialise().credential_type,
        data=b'\xde\xad\xbe\xef',
    )

    r = PdfFileReader(BytesIO(MINIMAL_AES256))
    with pytest.raises(
        misc.PdfReadError, match="Failed to deserialise password"
    ):
        r.security_handler.authenticate(cred_data)


def test_ser_deser_credential_unknown_cred_type():
    r = PdfFileReader(BytesIO(MINIMAL_AES256))
    r.decrypt("ownersecret")
    cred = r.security_handler.extract_credential()
    cred_data = SerialisedCredential(
        credential_type='foobar', data=cred.serialise().data
    )

    r = PdfFileReader(BytesIO(MINIMAL_AES256))
    with pytest.raises(
        misc.PdfReadError, match="credential type 'foobar' not known"
    ):
        r.security_handler.authenticate(cred_data)


@pytest.mark.parametrize('pw', ['usersecret', 'ownersecret'])
def test_ser_deser_credential_standard_sh_legacy(pw):
    out = _produce_legacy_encrypted_file(
        StandardSecuritySettingsRevision.RC4_OR_AES128, 16, True
    )
    r = PdfFileReader(out)
    r.decrypt(pw)
    cred = r.security_handler.extract_credential()
    assert cred['pwd_bytes'].native == pw.encode('utf8')
    assert cred['id1'].native is not None
    cred_data = cred.serialise()

    r = PdfFileReader(BytesIO(MINIMAL_AES256))
    result = r.security_handler.authenticate(cred_data)
    exp_status = AuthStatus.USER if pw.startswith('user') else AuthStatus.OWNER
    assert result.status == exp_status


@pytest.mark.parametrize('pw', ['usersecret', 'ownersecret'])
def test_ser_deser_credential_standard_sh_legacy_no_id1(pw):
    out = _produce_legacy_encrypted_file(
        StandardSecuritySettingsRevision.RC4_OR_AES128, 16, True
    )
    r = PdfFileReader(out)
    r.decrypt(pw)
    cred = r.security_handler.extract_credential()
    del cred['id1']
    cred_data = cred.serialise()

    r = PdfFileReader(out)
    with pytest.raises(misc.PdfReadError, match="id1"):
        r.security_handler.authenticate(cred_data)


def test_ser_deser_credential_standard_legacy_sh_extract_from_builder():
    sh = StandardSecurityHandler.build_from_pw_legacy(
        desired_owner_pass=b'ownersecret',
        desired_user_pass=b'usersecret',
        rev=StandardSecuritySettingsRevision.RC4_OR_AES128,
        keylen_bytes=16,
        id1=b'\xde\xad\xbe\xef',
    )
    cred = sh.extract_credential()
    assert cred['pwd_bytes'].native == b'ownersecret'
    assert cred['id1'].native == b'\xde\xad\xbe\xef'


def test_ser_deser_credential_pubkey():
    out = _produce_pubkey_encrypted_file(
        SecurityHandlerVersion.RC4_OR_AES128, 16, True, True
    )
    r = PdfFileReader(out)
    r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)
    cred_data = r.security_handler.extract_credential().serialise()

    r = PdfFileReader(out)
    result = r.security_handler.authenticate(cred_data)
    assert result.status == AuthStatus.USER


def test_ser_deser_credential_pubkey_sh_cannot_extract_from_builder():
    sh = PubKeySecurityHandler.build_from_certs(
        [PUBKEY_TEST_DECRYPTER.cert],
        keylen_bytes=16,
        version=SecurityHandlerVersion.RC4_OR_AES128,
        use_aes=True,
        use_crypt_filters=True,
        perms=PUBKEY_PERMS,
    )
    assert sh.extract_credential() is None


def test_ser_deser_credential_wrong_cred_type_pubkey():
    r = PdfFileReader(BytesIO(MINIMAL_AES256))
    r.decrypt("ownersecret")
    cred_data = r.security_handler.extract_credential().serialise()

    out = _produce_pubkey_encrypted_file(
        SecurityHandlerVersion.RC4_OR_AES128, 16, True, True
    )
    r = PdfFileReader(out)
    with pytest.raises(misc.PdfReadError, match="must be an instance of"):
        r.security_handler.authenticate(cred_data)


def test_ser_deser_credential_wrong_cred_type_standard():
    out = _produce_pubkey_encrypted_file(
        SecurityHandlerVersion.RC4_OR_AES128, 16, True, True
    )
    r = PdfFileReader(out)
    r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)
    cred_data = r.security_handler.extract_credential().serialise()

    r = PdfFileReader(BytesIO(MINIMAL_AES256))
    with pytest.raises(misc.PdfReadError, match="Standard auth.*must be a"):
        r.security_handler.authenticate(cred_data)


def test_ser_deser_credential_pubkey_corrupted():
    out = _produce_pubkey_encrypted_file(
        SecurityHandlerVersion.RC4_OR_AES128, 16, True, True
    )
    r = PdfFileReader(out)
    r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)
    cred = r.security_handler.extract_credential()
    cred_data = SerialisedCredential(
        credential_type=cred.serialise().credential_type,
        data=b'\xde\xad\xbe\xef',
    )

    r = PdfFileReader(out)
    with pytest.raises(
        misc.PdfReadError, match="Failed to decode serialised pubkey credential"
    ):
        r.security_handler.authenticate(cred_data)


def test_ser_deser_credential_wrong_cert():
    wrong_cert_cred_data = SimpleEnvelopeKeyDecrypter(
        cert=PUBKEY_SELFSIGNED_DECRYPTER.cert,
        private_key=PUBKEY_TEST_DECRYPTER.private_key,
    ).serialise()
    out = _produce_pubkey_encrypted_file(
        SecurityHandlerVersion.RC4_OR_AES128, 16, True, True
    )
    r = PdfFileReader(out)

    result = r.security_handler.authenticate(wrong_cert_cred_data)
    assert result.status == AuthStatus.FAILED


def test_ser_deser_credential_wrong_key():
    wrong_key_cred_data = SimpleEnvelopeKeyDecrypter(
        cert=PUBKEY_TEST_DECRYPTER.cert,
        private_key=PUBKEY_SELFSIGNED_DECRYPTER.private_key,
    ).serialise()
    out = _produce_pubkey_encrypted_file(
        SecurityHandlerVersion.RC4_OR_AES128, 16, True, True
    )
    r = PdfFileReader(out)

    # we're OK with this being an error, since a certificate match with a wrong
    # key is almost certainly indicative of something that shouldn't happen
    # in regular usage.
    with pytest.raises(misc.PdfReadError, match="envelope key"):
        r.security_handler.authenticate(wrong_key_cred_data)


@pytest.mark.parametrize('legacy', [True, False])
def test_encrypt_skipping_metadata(legacy):
    # we need to manually flag the metadata streams, since
    # pyHanko's PDF reader is (currently) not metadata-aware
    from pyhanko.pdf_utils.writer import copy_into_new_writer

    with open(
        os.path.join(PDF_DATA_DIR, "minimal-pdf-ua-and-a.pdf"), 'rb'
    ) as inf:
        w = copy_into_new_writer(PdfFileReader(inf))

    if legacy:
        sh = StandardSecurityHandler.build_from_pw_legacy(
            StandardSecuritySettingsRevision.RC4_OR_AES128,
            w._document_id[0].original_bytes,
            desired_owner_pass="secret",
            desired_user_pass="secret",
            keylen_bytes=16,
            use_aes128=True,
            perms=STD_PERMS,
            encrypt_metadata=False,
        )
        w._assign_security_handler(sh)
    else:
        w.encrypt("secret", "secret", encrypt_metadata=False)
    w.root['/Metadata'].apply_filter(
        "/Crypt", params={pdf_name("/Name"): pdf_name("/Identity")}
    )

    out = BytesIO()
    w.write(out)

    out.seek(0)
    r = PdfFileReader(out)
    mtd = r.root['/Metadata']
    assert not r.trailer['/Encrypt']['/EncryptMetadata']
    assert b'Test document' in mtd.encoded_data
    assert b'Test document' in mtd.data
    result = r.decrypt("secret")
    assert result.status == AuthStatus.OWNER

    assert r.trailer['/Info']['/Title'] == 'Test document'


def test_encrypt_skipping_metadata_pubkey():
    # we need to manually flag the metadata streams, since
    # pyHanko's PDF reader is (currently) not metadata-aware
    from pyhanko.pdf_utils.writer import copy_into_new_writer

    with open(
        os.path.join(PDF_DATA_DIR, "minimal-pdf-ua-and-a.pdf"), 'rb'
    ) as inf:
        w = copy_into_new_writer(PdfFileReader(inf))

    w.encrypt_pubkey([PUBKEY_TEST_DECRYPTER.cert], encrypt_metadata=False)
    w.root['/Metadata'].apply_filter(
        "/Crypt", params={pdf_name("/Name"): pdf_name("/Identity")}
    )

    out = BytesIO()
    w.write(out)

    out.seek(0)
    r = PdfFileReader(out)
    mtd = r.root['/Metadata']
    assert b'Test document' in mtd.encoded_data
    assert b'Test document' in mtd.data
    result = r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)
    assert result.status == AuthStatus.USER

    assert r.trailer['/Info']['/Title'] == 'Test document'


def test_pubkey_rc4_envelope():
    fname = os.path.join(PDF_DATA_DIR, "minimal-pubkey-rc4-envelope.pdf")
    with open(fname, 'rb') as inf:
        r = PdfFileReader(inf)
        result = r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER_OLD)
        assert result.status == AuthStatus.USER
        assert b'Hello' in r.root['/Pages']['/Kids'][0]['/Contents'].data


def test_unknown_envelope_enc_type():
    fname = os.path.join(
        PDF_DATA_DIR, "minimal-pubkey-unknown-envelope-alg.pdf"
    )
    with open(fname, 'rb') as inf:
        r = PdfFileReader(inf)
        with pytest.raises(misc.PdfError, match="Cipher.*not allowed"):
            r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER_OLD)


BASIC_R6_ENC_DICT = generic.DictionaryObject(
    {
        pdf_name('/Filter'): pdf_name('/Standard'),
        pdf_name('/O'): generic.ByteStringObject(
            binascii.unhexlify(
                "047761f7f568bfacb096382f2fc7cc94ffd87f33dc472ca4"
                "a3a3a78c739c77df26a794a7819aff59b3b85780c0fafe9f"
            )
        ),
        pdf_name('/U'): generic.ByteStringObject(
            binascii.unhexlify(
                "446ea469061c56060b56fa0296bfd32cc54fa9175e92ef5c"
                "0b945f4c810e309a03af4a2ff103bbd4db065e036f78ac4c"
            )
        ),
        pdf_name('/OE'): generic.ByteStringObject(
            binascii.unhexlify(
                "65012afdd09b34431117b9fa5f557202"
                "940dece9758d53c61fc5ff436cf2515c"
            )
        ),
        pdf_name('/UE'): generic.ByteStringObject(
            binascii.unhexlify(
                "71071c8c117abc19d26b5efc44a08066"
                "7ef3c3665ad3bc8f5f5b58126a15d931"
            )
        ),
        pdf_name('/Perms'): generic.ByteStringObject(
            binascii.unhexlify("b8729f735b0976d80f61d16bcfe09273")
        ),
        pdf_name('/P'): generic.NumberObject(-4),
        pdf_name('/V'): generic.NumberObject(5),
        pdf_name('/R'): generic.NumberObject(6),
        pdf_name('/Length'): generic.NumberObject(256),
        pdf_name('/EncryptMetadata'): generic.BooleanObject(True),
        pdf_name('/StmF'): pdf_name('/StdCF'),
        pdf_name('/StrF'): pdf_name('/StdCF'),
        pdf_name('/CF'): generic.DictionaryObject(
            {
                pdf_name('/StdCF'): generic.DictionaryObject(
                    {
                        pdf_name('/AuthEvent'): pdf_name('/DocOpen'),
                        pdf_name('/CFM'): pdf_name('/AESV3'),
                        pdf_name('/Length'): generic.NumberObject(32),
                    }
                )
            }
        ),
    }
)


@pytest.mark.parametrize(
    'enc_entry,delete,err',
    [
        ('/OE', False, "be 32 bytes long"),
        ('/OE', True, "be 32 bytes long"),
        ('/UE', False, "be 32 bytes long"),
        ('/UE', True, "be 32 bytes long"),
        ('/O', False, "be 48 bytes long"),
        ('/O', True, "be present"),
        ('/U', False, "be 48 bytes long"),
        ('/U', True, "be present"),
        ('/Perms', False, "be 16 bytes long"),
        ('/Perms', True, "be 16 bytes long"),
    ],
)
def test_r6_values(enc_entry, delete, err):
    enc_dict = generic.DictionaryObject(BASIC_R6_ENC_DICT)
    if delete:
        del enc_dict[enc_entry]
    else:
        enc_dict[enc_entry] = generic.ByteStringObject(b'\xde\xad\xbe\xef')
    with pytest.raises(misc.PdfError, match=err):
        StandardSecurityHandler.instantiate_from_pdf_object(enc_dict)


@pytest.mark.parametrize('entry', ["/U", "/O"])
def test_legacy_o_u_values(entry):
    r = PdfFileReader(BytesIO(VECTOR_IMAGE_PDF))
    w = writer.PdfFileWriter()
    sh = StandardSecurityHandler.build_from_pw_legacy(
        StandardSecuritySettingsRevision.RC4_OR_AES128,
        w._document_id[0].original_bytes,
        "ownersecret",
        "usersecret",
        keylen_bytes=True,
        use_aes128=True,
        perms=STD_PERMS,
    )
    w.security_handler = sh
    enc_dict = sh.as_pdf_object()
    enc_dict[entry] = generic.ByteStringObject(b"\xde\xad\xbe\xef")
    w._encrypt = w.add_object(enc_dict)
    new_page_tree = w.import_object(
        r.root.raw_get('/Pages'),
    )
    w.root['/Pages'] = new_page_tree
    out = BytesIO()
    w.write(out)

    with pytest.raises(misc.PdfError, match="be 32 bytes long"):
        # noinspection PyStatementEffect
        PdfFileReader(out).root['/Pages']['/Kids'][0]['/Content'].data


def test_key_length_constraint():
    enc_dict = generic.DictionaryObject(BASIC_R6_ENC_DICT)
    enc_dict['/Length'] = generic.NumberObject(333)
    with pytest.raises(misc.PdfError, match="must be a multiple of 8"):
        StandardSecurityHandler.instantiate_from_pdf_object(enc_dict)


def test_perms_decrypt_bogus():
    enc_dict = generic.DictionaryObject(BASIC_R6_ENC_DICT)
    enc_dict['/Perms'] = generic.ByteStringObject(b'\xde\xad\xbe\xef' * 4)
    sh = StandardSecurityHandler.instantiate_from_pdf_object(enc_dict)
    with pytest.raises(misc.PdfError, match="tampered"):
        sh.authenticate("usersecret")


def test_legacy_no_r6():
    with pytest.raises(ValueError, match="not supported"):
        _produce_legacy_encrypted_file(
            StandardSecuritySettingsRevision.AES256, 32, True
        )


def test_legacy_cf_req():
    with pytest.raises(misc.PdfError, match="s5 is required"):
        PubKeySecurityHandler.build_from_certs(
            [PUBKEY_TEST_DECRYPTER.cert],
            keylen_bytes=16,
            version=SecurityHandlerVersion.RC4_OR_AES128,
            use_aes=True,
            use_crypt_filters=False,
        )


def test_add_recp_before_auth_fail():
    out = _produce_pubkey_encrypted_file(
        SecurityHandlerVersion.RC4_OR_AES128, 16, True, True
    )
    r = PdfFileReader(out)
    cf = r.security_handler.get_stream_filter()
    with pytest.raises(misc.PdfError, match="before authenticating"):
        cf.add_recipients(
            [PUBKEY_SELFSIGNED_DECRYPTER.cert],
            policy=pubkey.RecipientEncryptionPolicy(),
        )


def test_add_recp_after_key_deriv():
    out = _produce_pubkey_encrypted_file(
        SecurityHandlerVersion.RC4_OR_AES128, 16, True, True
    )
    r = PdfFileReader(out)
    r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)
    cf = r.security_handler.get_stream_filter()
    assert cf.shared_key is not None
    with pytest.raises(misc.PdfError, match="after deriving.*shared key"):
        cf.add_recipients(
            [PUBKEY_SELFSIGNED_DECRYPTER.cert],
            policy=pubkey.RecipientEncryptionPolicy(),
        )


def test_encrypted_obj_stm():
    r = PdfFileReader(BytesIO(MINIMAL))
    w = writer.copy_into_new_writer(r, writer_kwargs={'stream_xrefs': True})
    obj_stm = w.prepare_object_stream()
    objref = w.add_object(
        generic.TextStringObject("Hello there"), obj_stream=obj_stm
    )
    w.encrypt("ownersecret", "usersecret")
    out = BytesIO()
    w.write(out)
    r = PdfFileReader(out)
    result = r.decrypt("ownersecret")
    assert result.status == AuthStatus.OWNER

    # assert that the content was present in the actual content stream
    # in unencrypted form (after decrypting the wrapping object stream)
    new_objref = generic.Reference(objref.idnum, objref.generation, pdf=r)
    xref_data = r.xrefs[new_objref]
    stm = r.get_object(
        generic.Reference(xref_data.obj_stream_id, pdf=r),
        transparent_decrypt=False,
    )
    assert b"(Hello there)" in stm.decrypted.data

    # assert that the object can be correctly retrieved
    assert new_objref.get_object() == "Hello there"


def test_add_crypt_filter_to_stream_without_security_handler():
    dummy_stream = generic.StreamObject(stream_data=b"1001")
    with pytest.raises(misc.PdfStreamError, match="no security handler"):
        dummy_stream.add_crypt_filter()


@pytest.mark.parametrize(
    "fname,strict",
    [
        ("malformed-encrypt-dict1.pdf", True),
        ("malformed-encrypt-dict2.pdf", True),
        ("malformed-encrypt-dict2.pdf", False),
    ],
)
def test_malformed_crypt(fname, strict):
    with open(os.path.join(PDF_DATA_DIR, fname), 'rb') as inf:
        r = PdfFileReader(inf, strict=strict)
        with pytest.raises(misc.PdfReadError, match='Encryption settings'):
            r.encrypt_dict


def test_tolerate_direct_encryption_dict_in_nonstrict():
    fname = 'malformed-encrypt-dict1.pdf'
    with open(os.path.join(PDF_DATA_DIR, fname), 'rb') as inf:
        r = PdfFileReader(inf, strict=False)
        r.decrypt('ownersecret')
        data = r.root['/Pages']['/Kids'][0]['/Contents'].data
        assert b'Hello' in data


def test_gcm_standard():
    w = writer.copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL)))

    sh = StandardSecurityHandler.build_from_pw(
        "secret", pdf_mac=False, use_gcm=True
    )
    w._assign_security_handler(sh)
    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    r.decrypt("secret")
    page_content = r.root['/Pages']['/Kids'][0]['/Contents'].data
    assert b"Hello" in page_content

    iso_exts = {
        int(ext.get_object()['/ExtensionLevel'])
        for ext in r.root['/Extensions']['/ISO_']
    }
    assert iso_exts == {32003}


def _gcm_standard_tamper(tamperer):
    w = writer.copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL)))

    sh = StandardSecurityHandler.build_from_pw(
        "secret", pdf_mac=False, use_gcm=True
    )
    w._assign_security_handler(sh)
    out = BytesIO()
    w.write(out)

    class NeverDecryptReader(PdfFileReader):
        def __init__(self):
            super().__init__(out)

        @property
        def security_handler(self):
            return None

    r = NeverDecryptReader()
    w = IncrementalPdfFileWriter.from_reader(r)
    page_dict = w.root['/Pages']['/Kids'][0]
    content: generic.StreamObject = page_dict['/Contents']
    content._encoded_data = tamperer(content.encoded_data)
    w.update_container(content)
    w._update_meta = lambda: None
    w.write_in_place()

    r = PdfFileReader(out)
    r.decrypt("secret")

    # this should work
    assert "https" in r.root['/Extensions']['/ISO_'][0]['/URL']

    # this shouldn't
    with pytest.raises(misc.PdfReadError, match="Invalid GCM tag"):
        len(r.root['/Pages']['/Kids'][0]['/Contents'].data)


def test_gcm_change_content():
    def tamper(ciphertext):
        out = BytesIO()
        out.write(ciphertext)
        out.seek(14)
        out.write(b"\xde\xad\xbe\xef")
        return out.getvalue()

    _gcm_standard_tamper(tamper)


def test_gcm_remove_tag():
    def tamper(ciphertext):
        return ciphertext[:-16]

    _gcm_standard_tamper(tamper)


def test_gcm_change_nonce():
    def tamper(ciphertext):
        out = BytesIO()
        out.write(ciphertext)
        out.seek(0)
        out.write(bytes(12))
        return out.getvalue()

    _gcm_standard_tamper(tamper)


def test_gcm_pubkey():
    w = writer.copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL)))

    sh = PubKeySecurityHandler.build_from_certs(
        [PUBKEY_TEST_DECRYPTER.cert],
        version=SecurityHandlerVersion.AES_GCM,
        pdf_mac=False,
    )
    w._assign_security_handler(sh)
    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)
    page_content = r.root['/Pages']['/Kids'][0]['/Contents'].data
    assert b"Hello" in page_content

    iso_exts = {
        int(ext.get_object()['/ExtensionLevel'])
        for ext in r.root['/Extensions']['/ISO_']
    }
    assert iso_exts == {32003}


def test_tolerate_empty_encrypted_string():
    with open(
        os.path.join(PDF_DATA_DIR, 'minimal-aes256-empty-encrypted-string.pdf'),
        'rb',
    ) as inf:
        r = PdfFileReader(inf)
        r.decrypt('secret')
        obj = r.root.raw_get('/Blah', decrypt=generic.EncryptedObjAccess.PROXY)
        assert isinstance(obj, generic.DecryptedObjectProxy)
        decrypted = obj.decrypted
        assert isinstance(
            decrypted, (generic.TextStringObject, generic.ByteStringObject)
        )
        assert decrypted.original_bytes == b""


def test_process_malformed_p_entry():
    with open(
        f'{PDF_DATA_DIR}/minimal-aes256-malformed-perms.pdf', 'rb'
    ) as inf:
        r = PdfFileReader(inf)
        with pytest.raises(
            misc.PdfReadError, match="Cannot parse.*as a permission"
        ):
            r.decrypt("usersecret")


def test_process_malformed_oe_entry():
    with open(f'{PDF_DATA_DIR}/minimal-aes256-malformed-oe.pdf', 'rb') as inf:
        r = PdfFileReader(inf)
        with pytest.raises(misc.PdfReadError, match="Expected string"):
            r.decrypt("usersecret")


@pytest.mark.parametrize(
    ['perm', 'expected_sint', 'expected_bytes'],
    (
        (StandardPermissions.allow_everything(), -4, b"\xff\xff\xff\xfc"),
        (STD_PERMS, -44, b"\xff\xff\xff\xd4"),
    ),
)
def test_std_permission_transformations(perm, expected_sint, expected_bytes):
    assert (perm.as_sint32(), perm.as_bytes()) == (
        expected_sint,
        expected_bytes,
    )


@pytest.mark.parametrize(
    ['perm', 'expected_bytes'],
    (
        (PubKeyPermissions.allow_everything(), b"\xff\xff\xff\xff"),
        (PUBKEY_PERMS, b"\xff\xff\xff\xd5"),
    ),
)
def test_pubkey_permission_transformations(perm, expected_bytes):
    assert perm.as_bytes() == expected_bytes


def test_pubkey_3des_decryption():
    with open(f"{PDF_DATA_DIR}/pubkey-3des-test.pdf", "rb") as inf:
        r = PdfFileReader(inf)
        result = r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)
        _validate_pubkey_decryption(r, result)


def test_pubkey_rc2_decryption():
    with open(f"{PDF_DATA_DIR}/pubkey-rc2-test.pdf", "rb") as inf:
        r = PdfFileReader(inf)
        result = r.decrypt_pubkey(PUBKEY_TEST_DECRYPTER)
        _validate_pubkey_decryption(r, result)



import logging
from typing import Optional, Tuple

import click
from pyhanko.sign import fields

logger = logging.getLogger("cli")


def _warn_empty_passphrase():
    click.echo(
        click.style(
            "WARNING: passphrase is empty. If you intended to use an "
            "unencrypted private key, use --no-pass instead.",
            bold=True,
        )
    )


readable_file = click.Path(exists=True, readable=True, dir_okay=False)
writable_file = click.Path(writable=True, dir_okay=False)


def _index_page(page):
    try:
        page_ix = int(page)
        if not page_ix:
            raise ValueError
        if page_ix > 0:
            # subtract 1 from the total, since that's what people expect
            # when referring to a page index
            return page_ix - 1
        else:
            # keep negative indexes as-is.
            return page_ix
    except ValueError:
        raise click.ClickException(
            "Sig field parameter PAGE should be a nonzero integer, "
            "not %s." % page
        )


def parse_field_location_spec(
    spec: str, require_full_spec: bool = True
) -> Tuple[str, Optional[fields.SigFieldSpec]]:
    try:
        page, box, name = spec.split('/')
    except ValueError:
        if require_full_spec:
            raise click.ClickException(
                "Sig field spec should be of the form PAGE/X1,Y1,X2,Y2/NAME."
            )
        else:
            # interpret the entire string as a field name
            return spec, None

    page_ix = _index_page(page)

    try:
        x1, y1, x2, y2 = map(int, box.split(','))
    except ValueError:
        raise click.ClickException(
            "Sig field parameters X1,Y1,X2,Y2 should be four integers."
        )

    return name, fields.SigFieldSpec(
        sig_field_name=name, on_page=page_ix, box=(x1, y1, x2, y2)
    )



from io import BytesIO

import pytest
from certomancer.integrations.illusionist import Illusionist
from certomancer.registry import ArchLabel, CertLabel
from freezegun import freeze_time
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko.sign import signers
from pyhanko.sign.ades.api import CAdESSignedAttrSpec, SignerAttrSpec
from pyhanko.sign.ades.cades_asn1 import (
    CommitmentTypeIndication,
    SignaturePolicyIdentifier,
)
from pyhanko.sign.validation import ades
from pyhanko.sign.validation.policy_decl import (
    PdfSignatureValidationSpec,
    SignatureValidationSpec,
)

from pyhanko_certvalidator.context import CertValidationPolicySpec
from pyhanko_certvalidator.registry import (
    SimpleCertificateStore,
    SimpleTrustManager,
)

from .samples import CERTOMANCER, MINIMAL_ONE_FIELD, SAMPLE_GROUP_ATTR
from .signing_commons import (
    DUMMY_POLICY_ID,
    DUMMY_TS,
    FROM_CA,
    FROM_ECC_CA,
    live_testing_vc,
)
from .test_ades_validation import DEFAULT_SIG_VALIDATION_SPEC
from .test_pades import PADES


async def _generate_basic_report(requests_mock, out):
    with freeze_time('2020-11-25'):
        r = PdfFileReader(out)
        live_testing_vc(requests_mock)
        result = await ades.ades_basic_validation(
            r.embedded_signatures[0].signed_data,
            validation_spec=DEFAULT_SIG_VALIDATION_SPEC,
            raw_digest=r.embedded_signatures[0].compute_digest(),
        )
        from pyhanko.sign.validation.report.tools import generate_report

        return generate_report(r.embedded_signatures[0], result)


@pytest.mark.asyncio
async def test_pades_basic_report_smoke_test(requests_mock):
    with freeze_time('2020-11-20'):
        w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
        out = await signers.async_sign_pdf(
            w,
            signers.PdfSignatureMetadata(field_name='Sig1', subfilter=PADES),
            signer=FROM_CA,
            timestamper=DUMMY_TS,
        )
    report = await _generate_basic_report(requests_mock, out)
    assert 'urn:etsi:019102:mainindication:total-passed' in report


@pytest.mark.asyncio
async def test_pades_basic_failing_report_smoke_test(requests_mock):
    with freeze_time('2020-11-20'):
        w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
        out = await signers.async_sign_pdf(
            w,
            signers.PdfSignatureMetadata(field_name='Sig1', subfilter=PADES),
            signer=FROM_CA,
            timestamper=DUMMY_TS,
        )
        out.seek(10)
        out.write(b'@')

    report = await _generate_basic_report(requests_mock, out)
    assert 'urn:etsi:019102:mainindication:total-failed' in report


@pytest.mark.asyncio
async def test_pades_basic_indeteriminate_report_smoke_test(requests_mock):
    with freeze_time('2020-11-20'):
        w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
        out = await signers.async_sign_pdf(
            w,
            signers.PdfSignatureMetadata(field_name='Sig1', subfilter=PADES),
            signer=FROM_ECC_CA,
            timestamper=DUMMY_TS,
        )

    report = await _generate_basic_report(requests_mock, out)
    assert 'urn:etsi:019102:mainindication:indeterminate' in report


async def _generate_lta_report(
    requests_mock, out, policy=DEFAULT_SIG_VALIDATION_SPEC
):
    with freeze_time('2028-11-25'):
        r = PdfFileReader(out)
        live_testing_vc(requests_mock)
        result = await ades.ades_lta_validation(
            r.embedded_signatures[0],
            pdf_validation_spec=PdfSignatureValidationSpec(
                signature_validation_spec=policy
            ),
        )
        from pyhanko.sign.validation.report.tools import generate_report

        return generate_report(r.embedded_signatures[0], result)


@pytest.mark.asyncio
async def test_pades_lta_report_smoke_test(requests_mock):
    with freeze_time('2020-11-20'):
        w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
        out = await signers.async_sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                field_name='Sig1',
                subfilter=PADES,
                use_pades_lta=True,
                validation_context=live_testing_vc(requests_mock),
            ),
            signer=FROM_CA,
            timestamper=DUMMY_TS,
        )
    report = await _generate_lta_report(requests_mock, out)
    assert 'urn:etsi:019102:mainindication:total-passed' in report
    assert 'urn:etsi:019102:validationprocess:LTA' in report


@pytest.mark.asyncio
async def test_pades_with_attributes_report_smoke_test(requests_mock):
    pki_arch = CERTOMANCER.get_pki_arch(ArchLabel('testing-ca-with-aa'))

    authorities = [
        pki_arch.get_cert('root'),
        pki_arch.get_cert('interm'),
        pki_arch.get_cert('root-aa'),
        pki_arch.get_cert('interm-aa'),
        pki_arch.get_cert('leaf-aa'),
    ]
    sig_validation_spec = SignatureValidationSpec(
        cert_validation_policy=CertValidationPolicySpec(
            trust_manager=SimpleTrustManager.build([pki_arch.get_cert('root')]),
            revinfo_policy=(
                DEFAULT_SIG_VALIDATION_SPEC.cert_validation_policy.revinfo_policy
            ),
        ),
        ac_validation_policy=CertValidationPolicySpec(
            trust_manager=SimpleTrustManager.build(
                [pki_arch.get_cert('root-aa')]
            ),
            revinfo_policy=(
                DEFAULT_SIG_VALIDATION_SPEC.cert_validation_policy.revinfo_policy
            ),
        ),
    )
    signer = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=SimpleCertificateStore.from_certs(authorities),
    )
    Illusionist(pki_arch).register(requests_mock)
    with freeze_time('2020-11-20'):
        w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
        out = await signers.async_sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                field_name='Sig1',
                subfilter=PADES,
                use_pades_lta=True,
                validation_context=live_testing_vc(requests_mock),
                cades_signed_attr_spec=CAdESSignedAttrSpec(
                    commitment_type=CommitmentTypeIndication(
                        {'commitment_type_id': 'proof_of_creation'}
                    ),
                    timestamp_content=True,
                    signature_policy_identifier=SignaturePolicyIdentifier(
                        {'signature_policy_id': DUMMY_POLICY_ID}
                    ),
                    signer_attributes=SignerAttrSpec(
                        claimed_attrs=[SAMPLE_GROUP_ATTR],
                        certified_attrs=[
                            pki_arch.get_attr_cert(
                                CertLabel('alice-role-with-rev')
                            )
                        ],
                    ),
                ),
            ),
            signer=signer,
            timestamper=DUMMY_TS,
        )
    report = await _generate_lta_report(
        requests_mock, out, policy=sig_validation_spec
    )
    assert 'urn:etsi:019102:mainindication:total-passed' in report
    assert 'claimed' in report
    assert 'certified' in report
    assert 'urn:etsi:019102:validationprocess:LTA' in report
